#ifndef SLIP_ITERATOR_H
#define SLIP_ITERATOR_H

// Copyright (C) 2022 Arthur I. Schwarz
//
// This file is part of the C++ SLIP library.  This library is free
// software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the
// Free Software Foundation; either version 3, or (at your option)
// any later version.

// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// Under Section 7 of GPL version 3, you are granted additional
// permissions described in the GCC Runtime Library Exception, version
// 3.1, as published by the Free Software Foundation.

// You should have received a copy of the GNU General Public License and
// a copy of the GCC Runtime Library Exception along with this program;
// see the files COPYING3 and COPYING.RUNTIME respectively.  If not, see
// <http://www.gnu.org/licenses/>.

/**
 * @file   SlipIterator.h
 * @author A. Schwarz
 * @date   Aug 24, 2022
 */

# include "SlipCell.h"
# include "SlipCellBase.h"
# include "SlipDatum.h"
# include "SlipErr.h"
# include "SlipException.h"
# include "SlipHeader.h"
# include "SlipSublist.h"

namespace slip {

/**
*  @brief Iterator base class.
* 
*  This base class contains operations, casts, and functions common
*  to the SlipSequencer and SlipReader. All operations in the SlipSequencer
*  are also in the SlipReader. Some operations in the SlipReader are
*  not available to the SlipSequencer.
* 
*  There are two iterators supported:
*  1. The SlipSequencer has no memory, as such, when a sublist is entered
*     there is no inbuilt mechanism to return to its enclosing list.
*     For purposes of this section, this type of iterator will be called
*     a "linear iterator". A linear iterator has light weight operations.
*  2. The SlipReader has memory, as such, when a sublist is entered there
*     is an inbuilt mechanism to return to the enclosing list. For purposes
*     of this section, this type of iterator will be called a "structured
*     iterator". A structured iterator has both light weight and heavy
*     weight operations. A structured iterator contains a reference to the
*     current list being traversed, as a result, a structured iterator can
*     participate in SlipHeader operations.
* 
*  A linear iterator is faster than a structured iterator when doing 
*  structured advances, i.e., advancing to a contained list, but can't
*  return to the list it is contained in. The speed of a structured and
*  linear iterator in linear advances should be the same.
* 
*  There a two aspects of an iterator.
*  1. Am iterator can be treated as a fundamental data type in the same manner
*     as SlipDatum. This includes casting operators, arithmetic and logical
*     operators, and bit manipulation.
*  2. The iterator can participate in some list operations. In this regard, some
*     operations are specific to the SlipReader iterator and the remaining
*     operations are shared by both the SlipSequencer and SlipReader.
* 
*  All functions descriptiions describe both iterators. All structured 
*  advances using a SlipSequencer can not return to the list left and 
*  has not knowledge of the current list header. All structured advances 
*  using a SlipReader can return to the list left and return a pointer 
*  to the current list being traversed.
* 
*  In addition, a SlipSequencer can be runtime stack based, that is,
*  does not need to allocate memory from memory. A SlipReader must
*  allocatate itself with the 'new' operator, and can not be allocated
*  on the stack.
* 
*  This means that:
*  <code><pre>
*     SlipSequencer  it;                          // is legal
*     SlipSequencer* it = new SlipSequencer(...); // is legal
*     SlipReader     it;                          // is not legal
*     SlipReader*    it = new SlipReader(...);    // is legal
*  </pre></code>
*/
   class SlipIterator { 

            /*****************************************
             *               Iteration               *
            ******************************************/
/**
 * @fn SlipCell& SlipIterator::advanceLEL()
 * @brief Advance the iterator linearly to the previous SlipDatum cell.
 *
 * The advance will skip over all non-SlipDatum cells, and will either
 * terminate on a SlipDatum cell or on the list header. If the initial
 * state is on the list header, then the immediately preceding cell (the)
 * bottom of the list) will be referenced. If there are no SiipDatum
 * cells then this advance will always terminate on the header.
 */

 /**
 * @fn SlipCell& SlipIterator::advanceLER()
 * @brief Advance the iterator linearly to the next SlipDatum cell.
 *
 * The advance will skip over all non-SlipDatum cells, and will either
 * terminate on a SlipDatum cell or on the list header. If the initial
 * state is on the list header, then the immediately next cell (the
 * top of the list) will be referenced. If there are no SiipDatum
 * cells then this advance will always terminate on the header.
 */

 /**
 * @fn SlipCell& SlipIterator::advanceLNL()
 * @brief Advance the iterator linearly to the previous SlipSublist cell.
 *
 * The advance will skip over all non-SlipSublist cells, and will either
 * terminate on a SlipSublist cell or on the list header. If the initial
 * state is on the list header, then the immediately preceding SlipSublist
 * cell (the last SlipSublist Cell) will be referenced. If there are no
 * SlipSublist cells then this advance will always terminate on the header.
 */

 /**
 * @fn SlipCell& SlipIterator::advanceLNR()
 * @brief Advance the iterator linearly to the next SlipSublist cell.
 *
 * The advance will skip over all non-SlipSublist cells, and will either
 * terminate on a SlipSublist cell or on the list header. If the initial
 * state is on the list header, then the immediately next SlipSublist
 * cell (the first SlipSublist Cell) will be referenced_. If there are no
 * SlipSublist cells then this advance will always terminate on the header.
 */

 /**
 * @fn SlipCell& SlipIterator::advanceLWL()
 * @brief Advance the iterator linearly to the previous SlipDatum or SlipSublist cell.
 *
 * The advance will advance to the previous Slip cell. If the initial
 * state is on the list header, then the immediately preceding cell (the
 * bottom of the list) will be referenced. If there are no SiipDatum
 * cells then this advance will always terminate on the header.
 */

 /**
 * @fn SlipCell& SlipIterator::advanceLWR()
 * @brief Advance the iterator linearly to the next SlipDatum or SlipSublist cell.
 *
 * The advance will skip over all non-SlipDatum cells, and will either
 * terminate on a SlipDatum cell or on the list header. If the initial
 * state is on the list header, then the immediately preceding cell (the
 * bottom of the list) will be referenced. If there the list is empty
 * then the advance will be to the list header.
 */

 /**
 * @fn SlipCell& SlipIterator::advanceSEL()
 * @brief Advance the iterator structurally to the previous SlipDatum cell.
 *
 * If a SlipSublist cell is seen, then the advance will 'enter' the
 * sublist and advance to the SlipDatum cell preceding the list header.
 * If a SlipSublist cell is seen before a SlipDAtum cell is seen, then
 * the advance will 'enter' the new list. If the list is empty, the
 * advance will remain on the list header.
 *
 * SlipSequencers can not return from a sublist. SlipReader can.
 */

 /**
 * @fn SlipCell& SlipIterator::advanceSER()
 * @brief Advance the iterator structurally to the next SlipDatum cell.
 *
 * If a SlipSublist cell is seen, then the advance will 'enter' the
 * sublist and advance to the next SlipDatum cell following the list header.
 * If a SlipSublist cell is seen before a SlipDatum cell is seen, then
 * the advance will 'enter' the new list. If the list is empty, the
 * advance will remain on the list header.
 *
 * SlipSequencers can not return from a sublist. SlipReader can.
 */

 /**
 * @fn SlipCell& SlipIterator::advanceSNL()
 * @brief Advance the iterator structurally to the previous SlipSublist cell.
 *
 * If the iterator is not on a SlipSublist cell, then the advance will
 * be to the preceding SlipSublist cell, or to the list header if none
 * exist. If the iterator is on a SlipSublist cell then the advance will
 * enter the new list and terminate on the SlipSublist cell immediately
 * preceding the list header, or if none exist, will terminate on the
 * list header.
 *
 * SlipSequencers can not return from a sublist. SlipReader can.
 */

 /**
 * @fn SlipCell& SlipIterator::advanceSNR()
 * @brief Advance the iterator linearly to the next SlipSublist cell.
 *
 * If the iterator is not on a SlipSublist cell, then the advance will
 * be to the next SlipSublist cell, or to the list header if none
 * exist. If the iterator is on a SlipSublist cell then the advance will
 * enter the new list and terminate on the SlipSublist cell immediately
 * following the list header, or if none exist, will terminate on the
 * list header.
 *
 * SlipSequencers can not return from a sublist. SlipReader can.
 */

 /**
  * @fn SlipCell& SlipIterator::advanceSWL()
  * @brief Advance the iterator linearly to the previous SlipDatum or SlipSublist cell.
  *
  * If the iterator is not on a Slip cell, then the advance will
  * be to the next SlipSublist cell, or to the list header if none
  * exist. If the iterator is on a SlipSublist cell then the advance will
  * enter the new list and terminate on the SlipSublist cell immediately
  * preceding the list header, or if none exist, will terminate on the
  * list header.
  *
  * SlipSequencers can not return from a sublist. SlipReader can.
  */

  /**
  * @fn SlipCell& SlipIterator::advanceSWR()
  * @brief Advance the iterator linearly to the next SlipDatum or SlipSublist cell.
  *
  * If the iterator is not on a Slip cell, then the advance will
  * be to the next cell, or to the list header if none
  * exist. If the iterator is on a SlipSublist cell then the advance will
  * enter the new list and terminate on the SlipSublist cell immediately
  * following the list header, or if none exist, will terminate on the
  * list header.
  *
  * SlipSequencers can not return from a sublist. SlipReader can.
  */
   private:
   protected:

   /*********************************************************
    *                   Public Functions                    *
   **********************************************************/
   public:
//      virtual ~SlipIterator() = 0;

   /*****************************************
    *               Iteration               *
   ******************************************/
      virtual SlipCell& advanceLEL() = 0;                               // advance linear element left
      virtual SlipCell& advanceLER() = 0;                               // advance linear element right
      virtual SlipCell& advanceLNL() = 0;                               // advance linear name left
      virtual SlipCell& advanceLNR() = 0;                               // advance linear name right
      virtual SlipCell& advanceLWL() = 0;                               // advance linear word right
      virtual SlipCell& advanceLWR() = 0;                               // advance linear word right
      virtual SlipCell& advanceSEL() = 0;                               // advance structural element left
      virtual SlipCell& advanceSER() = 0;                               // advance structural element right
      virtual SlipCell& advanceSNL() = 0;                               // advance structural name left
      virtual SlipCell& advanceSNR() = 0;                               // advance structural name right
      virtual SlipCell& advanceSWL() = 0;                               // advance structural word right
      virtual SlipCell& advanceSWR() = 0;                               // advance structural word right

   /*********************************************************
   *                  Casting Operators                    *
   *********************************************************/
      operator bool()   const { return (bool)(SlipDatum&)currentCell();   }
      operator UCHAR()  const { return (UCHAR)(SlipDatum&)currentCell();  }
      operator CHAR()   const { return (CHAR)(SlipDatum&)currentCell();   }
      operator ULONG()  const { return (ULONG)(SlipDatum&)currentCell();  }
      operator LONG()   const { return (LONG)(SlipDatum&)currentCell();   }
      operator DOUBLE() const { return (DOUBLE)(SlipDatum&)currentCell(); }
      operator PTR()    const { return (PTR)(SlipDatum&)currentCell();    }
      operator string() const { return *(string*)(SlipDatum&)currentCell(); }
      operator SlipSublist&() const { if (isSublist()) return (SlipSublist&)currentCell(); 
                                      else postError(__FILE__, __LINE__, SlipErr::E1044, "(SlipSublist&)", "", "SlipCell not a sublist.", currentCell()); 
                                      return (SlipSublist&)currentCell();}
      operator SlipHeader&() const { if (isHeader()) return (SlipHeader&)currentCell();
                                     else if (isSublist()) return ((SlipSublist&)currentCell()).getHeader();
                                     else postError(__FILE__, __LINE__, SlipErr::E1044, "(SlipHeader&)", "", "SlipCell not a header.", currentCell());
                                     return (SlipHeader&)currentCell();}

   /*************************************************
    *                Unary operators                *
   *************************************************/
      bool       operator!()     { return !(SlipDatum&)currentCell();  }
      ULONG      operator~()     { return ~(SlipDatum&)currentCell();  }
      SlipDatum  operator+()     { return +(SlipDatum&)currentCell();  }
      SlipDatum  operator-()     { return -(SlipDatum&)currentCell();  }
      SlipDatum& operator++()    { return ++(SlipDatum&)currentCell(); }
      SlipDatum& operator--()    { return --(SlipDatum&)currentCell(); }
      SlipDatum& operator++(int) { return ((SlipDatum&)currentCell())++; }
      SlipDatum& operator--(int) { return ((SlipDatum&)currentCell())--; }

   /*************************************************
    *                Binary operators               *
   *************************************************/
      SlipDatum  operator+(const SlipDatum& X) { return (SlipDatum&)currentCell() + X; }
      SlipDatum  operator-(const SlipDatum& X) { return (SlipDatum&)currentCell() - X; }
      SlipDatum  operator*(const SlipDatum& X) { return (SlipDatum&)currentCell() * X; }
      SlipDatum  operator/(const SlipDatum& X) { return (SlipDatum&)currentCell() / X; }
      LONG       operator%(const SlipDatum& X) { return (SlipDatum&)currentCell() % X; }

      SlipDatum  operator+(const DOUBLE     X) { return (SlipDatum&)currentCell() + X; }
      SlipDatum  operator-(const DOUBLE     X) { return (SlipDatum&)currentCell() - X; }
      SlipDatum  operator*(const DOUBLE     X) { return (SlipDatum&)currentCell() * X; }
      SlipDatum  operator/(const DOUBLE     X) { return (SlipDatum&)currentCell() / X; }

      SlipDatum  operator+(const bool   X) { return (SlipDatum&)currentCell() + X; }
      SlipDatum  operator-(const bool   X) { return (SlipDatum&)currentCell() - X; }
      SlipDatum  operator*(const bool   X) { return (SlipDatum&)currentCell() * X; }
      SlipDatum  operator/(const bool   X) { return (SlipDatum&)currentCell() / X; }
      LONG       operator%(const bool   X) { return (SlipDatum&)currentCell() % X; }

      SlipDatum  operator+(const UCHAR  X) { return (SlipDatum&)currentCell() + X; }
      SlipDatum  operator-(const UCHAR  X) { return (SlipDatum&)currentCell() - X; }
      SlipDatum  operator*(const UCHAR  X) { return (SlipDatum&)currentCell() * X; }
      SlipDatum  operator/(const UCHAR  X) { return (SlipDatum&)currentCell() / X; }
      LONG       operator%(const UCHAR  X) { return (SlipDatum&)currentCell() % X; }

      SlipDatum  operator+(const CHAR   X) { return (SlipDatum&)currentCell() + X; }
      SlipDatum  operator-(const CHAR   X) { return (SlipDatum&)currentCell() - X; }
      SlipDatum  operator*(const CHAR   X) { return (SlipDatum&)currentCell() * X; }
      SlipDatum  operator/(const CHAR   X) { return (SlipDatum&)currentCell() / X; }
      LONG       operator%(const CHAR   X) { return (SlipDatum&)currentCell() % X; }

      SlipDatum  operator+(const ULONG  X) { return (SlipDatum&)currentCell() + X; }
      SlipDatum  operator-(const ULONG  X) { return (SlipDatum&)currentCell() - X; }
      SlipDatum  operator*(const ULONG  X) { return (SlipDatum&)currentCell() * X; }
      SlipDatum  operator/(const ULONG  X) { return (SlipDatum&)currentCell() / X; }
      LONG       operator%(const ULONG  X) { return (SlipDatum&)currentCell() % X; }

      SlipDatum  operator+(const LONG   X) { return (SlipDatum&)currentCell() + X; }
      SlipDatum  operator-(const LONG   X) { return (SlipDatum&)currentCell() - X; }
      SlipDatum  operator*(const LONG   X) { return (SlipDatum&)currentCell() * X; }
      SlipDatum  operator/(const LONG   X) { return (SlipDatum&)currentCell() / X; }
      LONG       operator%(const LONG   X) { return (SlipDatum&)currentCell() % X; }

   /*************************************************
    *                Bit operators                  *
   *************************************************/
      LONG operator<<(const SlipDatum& X) { return (SlipDatum&)currentCell() << X; }
      LONG operator>>(const SlipDatum& X) { return (SlipDatum&)currentCell() >> X; }
      LONG operator&(const  SlipDatum& X) { return (SlipDatum&)currentCell()  & X; }
      LONG operator|(const  SlipDatum& X) { return (SlipDatum&)currentCell()  | X; }
      LONG operator^(const  SlipDatum& X) { return (SlipDatum&)currentCell()  ^ X; }

      LONG operator<<(const bool       X) { return (SlipDatum&)currentCell() << X; }
      LONG operator>>(const bool       X) { return (SlipDatum&)currentCell() >> X; }
      LONG operator& (const bool       X) { return (SlipDatum&)currentCell()  & X; }
      LONG operator| (const bool       X) { return (SlipDatum&)currentCell()  | X; }
      LONG operator^ (const bool       X) { return (SlipDatum&)currentCell()  ^ X; }

      LONG operator<<(const UCHAR      X) { return (SlipDatum&)currentCell() << X; }
      LONG operator>>(const UCHAR      X) { return (SlipDatum&)currentCell() >> X; }
      LONG operator& (const UCHAR      X) { return (SlipDatum&)currentCell()  & X; }
      LONG operator| (const UCHAR      X) { return (SlipDatum&)currentCell()  | X; }
      LONG operator^ (const UCHAR      X) { return (SlipDatum&)currentCell() ^ X; }

      LONG operator<<(const CHAR       X) { return (SlipDatum&)currentCell() << X; }
      LONG operator>>(const CHAR       X) { return (SlipDatum&)currentCell() >> X; }
      LONG operator& (const CHAR       X) { return (SlipDatum&)currentCell() & X; }
      LONG operator| (const CHAR       X) { return (SlipDatum&)currentCell() | X; }
      LONG operator^ (const CHAR       X) { return (SlipDatum&)currentCell() ^ X; }

      LONG operator<<(const ULONG      X) { return (SlipDatum&)currentCell() << X; }
      LONG operator>>(const ULONG      X) { return (SlipDatum&)currentCell() >> X; }
      LONG operator& (const ULONG      X) { return (SlipDatum&)currentCell()  & X; }
      LONG operator| (const ULONG      X) { return (SlipDatum&)currentCell()  | X; }
      LONG operator^ (const ULONG      X) { return (SlipDatum&)currentCell()  ^ X; }

      LONG operator<<(const LONG       X) { return (SlipDatum&)currentCell() << X; }
      LONG operator>>(const LONG       X) { return (SlipDatum&)currentCell() >> X; }
      LONG operator& (const LONG       X) { return (SlipDatum&)currentCell()  & X; }
      LONG operator| (const LONG       X) { return (SlipDatum&)currentCell()  | X; }
      LONG operator^ (const LONG       X) { return (SlipDatum&)currentCell()  ^ X; }

   /*************************************************
    *             Assignment operators              *
   *************************************************/
      SlipDatum& operator=(const SlipDatum& X) { return (SlipDatum&)operator=((SlipCell&)X); }
      SlipCell&  operator=(const SlipCell& X) { return (SlipDatum&)currentCell() = X; }
      SlipDatum& operator=(const bool    X)   { return (SlipDatum&)currentCell() = X; }
      SlipDatum& operator=(const CHAR    X)   { return (SlipDatum&)currentCell() = X; }
      SlipDatum& operator=(const DOUBLE  X)   { return (SlipDatum&)currentCell() = X; }
      SlipDatum& operator=(const LONG    X)   { return (SlipDatum&)currentCell() = X; }
      SlipDatum& operator=(const UCHAR   X)   { return (SlipDatum&)currentCell() = X; }
      SlipDatum& operator=(const ULONG   X)   { return (SlipDatum&)currentCell() = X; }
      SlipDatum& operator=(const PTR     X)   { return (SlipDatum&)currentCell() = X; }
      SlipDatum& operator=(const string& X)   { return (SlipDatum&)currentCell() = X; }
      SlipDatum& operator=(const string* X)   { return (SlipDatum&)currentCell() = X; }

   /*************************************************
    *        Compound Assignment operators          *
   *************************************************/
      SlipDatum& operator+= (const SlipDatum& X) { return (SlipDatum&)currentCell() +=  X; }
      SlipDatum& operator-= (const SlipDatum& X) { return (SlipDatum&)currentCell() -=  X; }
      SlipDatum& operator*= (const SlipDatum& X) { return (SlipDatum&)currentCell() *=  X; }
      SlipDatum& operator/= (const SlipDatum& X) { return (SlipDatum&)currentCell() /=  X; }
      SlipDatum& operator%= (const SlipDatum& X) { return (SlipDatum&)currentCell() %=  X; }
      SlipDatum& operator<<=(const SlipDatum& X) { return (SlipDatum&)currentCell() <<= X; }
      SlipDatum& operator>>=(const SlipDatum& X) { return (SlipDatum&)currentCell() >>= X; }
      SlipDatum& operator&= (const SlipDatum& X) { return (SlipDatum&)currentCell() &=  X; }
      SlipDatum& operator|= (const SlipDatum& X) { return (SlipDatum&)currentCell() |=  X; }
      SlipDatum& operator^= (const SlipDatum& X) { return (SlipDatum&)currentCell() ^=  X; }

      SlipDatum& operator+= (const DOUBLE     X) { return (SlipDatum&)currentCell() += X; }
      SlipDatum& operator-= (const DOUBLE     X) { return (SlipDatum&)currentCell() -= X; }
      SlipDatum& operator*= (const DOUBLE     X) { return (SlipDatum&)currentCell() *= X; }
      SlipDatum& operator/= (const DOUBLE     X) { return (SlipDatum&)currentCell() /= X; }

      SlipDatum& operator+= (const bool       X) { return (SlipDatum&)currentCell() += X; }
      SlipDatum& operator-= (const bool       X) { return (SlipDatum&)currentCell() -= X; }

      SlipDatum& operator+= (const UCHAR      X) { return (SlipDatum&)currentCell() += X;  }
      SlipDatum& operator-= (const UCHAR      X) { return (SlipDatum&)currentCell() -= X;  }
      SlipDatum& operator*= (const UCHAR      X) { return (SlipDatum&)currentCell() *= X;  }
      SlipDatum& operator/= (const UCHAR      X) { return (SlipDatum&)currentCell() /= X;  }
      SlipDatum& operator%= (const UCHAR      X) { return (SlipDatum&)currentCell() %= X;  }
      SlipDatum& operator<<=(const UCHAR      X) { return (SlipDatum&)currentCell() <<= X; }
      SlipDatum& operator>>=(const UCHAR      X) { return (SlipDatum&)currentCell() >>= X; }
      SlipDatum& operator&= (const UCHAR      X) { return (SlipDatum&)currentCell() &= X;  }
      SlipDatum& operator|= (const UCHAR      X) { return (SlipDatum&)currentCell() |= X;  }
      SlipDatum& operator^= (const UCHAR      X) { return (SlipDatum&)currentCell() ^= X;  }

      SlipDatum& operator+= (const CHAR       X) { return (SlipDatum&)currentCell() += X;  }
      SlipDatum& operator-= (const CHAR       X) { return (SlipDatum&)currentCell() -= X;  }
      SlipDatum& operator*= (const CHAR       X) { return (SlipDatum&)currentCell() *= X;  }
      SlipDatum& operator/= (const CHAR       X) { return (SlipDatum&)currentCell() /= X;  }
      SlipDatum& operator%= (const CHAR       X) { return (SlipDatum&)currentCell() %= X;  }
      SlipDatum& operator<<=(const CHAR       X) { return (SlipDatum&)currentCell() <<= X; }
      SlipDatum& operator>>=(const CHAR       X) { return (SlipDatum&)currentCell() >>= X; }
      SlipDatum& operator&= (const CHAR       X) { return (SlipDatum&)currentCell() &= X;  }
      SlipDatum& operator|= (const CHAR       X) { return (SlipDatum&)currentCell() |= X;  }
      SlipDatum& operator^= (const CHAR       X) { return (SlipDatum&)currentCell() ^= X;  }

      SlipDatum& operator+= (const ULONG      X) { return (SlipDatum&)currentCell() += X;  }
      SlipDatum& operator-= (const ULONG      X) { return (SlipDatum&)currentCell() -= X;  }
      SlipDatum& operator*= (const ULONG      X) { return (SlipDatum&)currentCell() *= X;  }
      SlipDatum& operator/= (const ULONG      X) { return (SlipDatum&)currentCell() /= X;  }
      SlipDatum& operator%= (const ULONG      X) { return (SlipDatum&)currentCell() %= X;  }
      SlipDatum& operator<<=(const ULONG      X) { return (SlipDatum&)currentCell() <<= X; }
      SlipDatum& operator>>=(const ULONG      X) { return (SlipDatum&)currentCell() >>= X; }
      SlipDatum& operator&= (const ULONG      X) { return (SlipDatum&)currentCell() &= X;  }
      SlipDatum& operator|= (const ULONG      X) { return (SlipDatum&)currentCell() |= X;  }
      SlipDatum& operator^= (const ULONG      X) { return (SlipDatum&)currentCell() ^= X;  }

      SlipDatum& operator+= (const LONG       X) { return (SlipDatum&)currentCell() += X;  }
      SlipDatum& operator-= (const LONG       X) { return (SlipDatum&)currentCell() += X;  }
      SlipDatum& operator*= (const LONG       X) { return (SlipDatum&)currentCell() *= X;  }
      SlipDatum& operator/= (const LONG       X) { return (SlipDatum&)currentCell() /= X;  }
      SlipDatum& operator%= (const LONG       X) { return (SlipDatum&)currentCell() %= X;  }
      SlipDatum& operator<<=(const LONG       X) { return (SlipDatum&)currentCell() <<= X; }
      SlipDatum& operator>>=(const LONG       X) { return (SlipDatum&)currentCell() >>= X; }
      SlipDatum& operator&= (const LONG       X) { return (SlipDatum&)currentCell() &= X;  }
      SlipDatum& operator|= (const LONG       X) { return (SlipDatum&)currentCell() |= X;  }
      SlipDatum& operator^= (const LONG       X) { return (SlipDatum&)currentCell() ^= X;  }

   /*************************************************
    *              Logical operators                *
   *************************************************/
      bool       operator==(const SlipIterator& X) { return (SlipDatum&)currentCell() == (SlipDatum&)X.currentCell(); }
      bool       operator!=(const SlipIterator& X) { return (SlipDatum&)currentCell() != (SlipDatum&)X.currentCell(); }
      bool       operator< (const SlipIterator& X) { return (SlipDatum&)currentCell()  < (SlipDatum&)X.currentCell(); }
      bool       operator<=(const SlipIterator& X) { return (SlipDatum&)currentCell() <= (SlipDatum&)X.currentCell(); }
      bool       operator>=(const SlipIterator& X) { return (SlipDatum&)currentCell() >= (SlipDatum&)X.currentCell(); }
      bool       operator> (const SlipIterator& X) { return (SlipDatum&)currentCell()  > (SlipDatum&)X.currentCell(); }
    
      bool       operator==(const SlipCell&  X) { return (SlipDatum&)currentCell() == (SlipDatum&)X; }
      bool       operator!=(const SlipCell&  X) { return (SlipDatum&)currentCell() != (SlipDatum&)X; }
      bool       operator< (const SlipCell&  X) { return (SlipDatum&)currentCell() <  (SlipDatum&)X; }
      bool       operator<=(const SlipCell&  X) { return (SlipDatum&)currentCell() <= (SlipDatum&)X; }
      bool       operator>=(const SlipCell&  X) { return (SlipDatum&)currentCell() >= (SlipDatum&)X; }
      bool       operator> (const SlipCell&  X) { return (SlipDatum&)currentCell() >  (SlipDatum&)X; }

      bool       operator!=(const bool      X) { return (SlipDatum&)currentCell() != X; }
      bool       operator< (const bool      X) { return (SlipDatum&)currentCell() < X;  }
      bool       operator<=(const bool      X) { return (SlipDatum&)currentCell() <= X; }
      bool       operator==(const bool      X) { return (SlipDatum&)currentCell() == X; }
      bool       operator>=(const bool      X) { return (SlipDatum&)currentCell() >= X; }
      bool       operator> (const bool      X) { return (SlipDatum&)currentCell() > X;  }

      bool       operator!=(const DOUBLE    X) { return (SlipDatum&)currentCell() != X; }
      bool       operator< (const DOUBLE    X) { return (SlipDatum&)currentCell() < X;  }
      bool       operator<=(const DOUBLE    X) { return (SlipDatum&)currentCell() <= X; }
      bool       operator==(const DOUBLE    X) { return (SlipDatum&)currentCell() == X; }
      bool       operator>=(const DOUBLE    X) { return (SlipDatum&)currentCell() >= X; }
      bool       operator> (const DOUBLE    X) { return (SlipDatum&)currentCell() > X;  }

      bool       operator!=(const UCHAR     X) { return (SlipDatum&)currentCell() != X; }
      bool       operator< (const UCHAR     X) { return (SlipDatum&)currentCell() < X;  }
      bool       operator<=(const UCHAR     X) { return (SlipDatum&)currentCell() <= X; }
      bool       operator==(const UCHAR     X) { return (SlipDatum&)currentCell() == X; }
      bool       operator>=(const UCHAR     X) { return (SlipDatum&)currentCell() >= X; }
      bool       operator> (const UCHAR     X) { return (SlipDatum&)currentCell() > X;  }

      bool       operator!=(const CHAR      X) { return (SlipDatum&)currentCell() != X; }
      bool       operator< (const CHAR      X) { return (SlipDatum&)currentCell() < X;  }
      bool       operator<=(const CHAR      X) { return (SlipDatum&)currentCell() <= X; }
      bool       operator==(const CHAR      X) { return (SlipDatum&)currentCell() == X; }
      bool       operator>=(const CHAR      X) { return (SlipDatum&)currentCell() >= X; }
      bool       operator> (const CHAR      X) { return (SlipDatum&)currentCell() > X;  }

      bool       operator!=(const ULONG     X) { return (SlipDatum&)currentCell() != X; }
      bool       operator< (const ULONG     X) { return (SlipDatum&)currentCell() < X;  }
      bool       operator<=(const ULONG     X) { return (SlipDatum&)currentCell() <= X; }
      bool       operator==(const ULONG     X) { return (SlipDatum&)currentCell() == X; }
      bool       operator>=(const ULONG     X) { return (SlipDatum&)currentCell() >= X; }
      bool       operator> (const ULONG     X) { return (SlipDatum&)currentCell() > X;  }

      bool       operator!=(const LONG      X) { return (SlipDatum&)currentCell() != X; }
      bool       operator< (const LONG      X) { return (SlipDatum&)currentCell() < X;  }
      bool       operator<=(const LONG      X) { return (SlipDatum&)currentCell() <= X; }
      bool       operator==(const LONG      X) { return (SlipDatum&)currentCell() == X; }
      bool       operator>=(const LONG      X) { return (SlipDatum&)currentCell() >= X; }
      bool       operator> (const LONG      X) { return (SlipDatum&)currentCell() > X;  }

      bool       operator!=(const PTR       X) { return (SlipDatum&)currentCell() != X; }
      bool       operator==(const PTR       X) { return (SlipDatum&)currentCell() == X; }

      bool       operator!=(const string&    X) { return (SlipDatum&)currentCell() != X; }
      bool       operator< (const string&    X) { return (SlipDatum&)currentCell() < X;  }
      bool       operator<=(const string&    X) { return (SlipDatum&)currentCell() <= X; }
      bool       operator==(const string&    X) { return (SlipDatum&)currentCell() == X; }
      bool       operator>=(const string&    X) { return (SlipDatum&)currentCell() >= X; }
      bool       operator> (const string&    X) { return (SlipDatum&)currentCell() > X;  }

      bool       operator!=(const string*    X) { return (SlipDatum&)currentCell() != X; }
      bool       operator< (const string*    X) { return (SlipDatum&)currentCell() < X;  }
      bool       operator<=(const string*    X) { return (SlipDatum&)currentCell() <= X; }
      bool       operator==(const string*    X) { return (SlipDatum&)currentCell() == X; }
      bool       operator>=(const string*    X) { return (SlipDatum&)currentCell() >= X; }
      bool       operator> (const string*    X) { return (SlipDatum&)currentCell() > X;  }

   /*****************************************
   *           Manipulating Cells          *
   ******************************************/
      SlipCell& check(SlipIterator& X) {
         SlipCell& cell = X.currentCell();
         if (!X.isHeader()) X.advanceLWL();
         return cell;
      }; // SlipCell& SlipIterator::insert(SlipIterator& X)

              // List Cells must be unlinked
      SlipCell& insLeft(SlipIterator& X)  { return currentCell().insLeft(check(X)); };                               // Insert X to current cell left
      SlipCell& insLeft(SlipCell&  X)     { return currentCell().insLeft(X); };
      SlipCell& insLeft(bool       X)     { return currentCell().insLeft(X); };
      SlipCell& insLeft(DOUBLE     X)     { return currentCell().insLeft(X); };
      SlipCell& insLeft(ULONG      X)     { return currentCell().insLeft(X); };
      SlipCell& insLeft(LONG       X)     { return currentCell().insLeft(X); };
      SlipCell& insLeft(UCHAR      X)     { return currentCell().insLeft(X); };
      SlipCell& insLeft(CHAR       X)     { return currentCell().insLeft(X); };
      SlipCell& insLeft(const PTR  X, const void* operation = ptrOP)  // Insert X to current cell left
                                      { return currentCell().insLeft(X, operation); };
      SlipCell& insLeft(const string& X, bool constFlag = false)      // Insert X to current cell left
                                      { return currentCell().insLeft(X, constFlag); };
      SlipCell& insLeft(const string* X, bool constFlag = false)      // Insert X to current cell left
                                      { return currentCell().insLeft(X, constFlag); };

         // List Cells must be unlinked
      SlipCell& insRight(SlipIterator& X) { return currentCell().insRight(check(X)); };
      SlipCell& insRight(SlipCell&  X)    { return currentCell().insRight(X); };
      SlipCell& insRight(bool       X)    { return currentCell().insRight(X); };
      SlipCell& insRight(DOUBLE     X)    { return currentCell().insRight(X); };
      SlipCell& insRight(ULONG      X)    { return currentCell().insRight(X); };
      SlipCell& insRight(LONG       X)    { return currentCell().insRight(X); };
      SlipCell& insRight(UCHAR      X)    { return currentCell().insRight(X); };
      SlipCell& insRight(CHAR       X)    { return currentCell().insRight(X); };
      SlipCell& insRight(const PTR  X, const void* operation = ptrOP)  // Insert X to current cell Right
                                       { return currentCell().insRight(X, operation); };
      SlipCell& insRight(const string& X, bool constFlag = false)      // Insert X to current cell Right
                                       { return currentCell().insRight(X, constFlag); };
      SlipCell& insRight(const string* X, bool constFlag = false)      // Insert X to current cell Right
                                       { return currentCell().insRight(X, constFlag); };

      SlipCell& moveLeft(SlipCell& X)     { currentCell().moveLeft(X); return currentCell(); }
      SlipCell& moveRight(SlipCell& X)    { currentCell().moveRight(X); return currentCell(); }
      SlipCell& moveLeft(SlipIterator& X) { currentCell().moveLeft(check(X)); return currentCell(); }
      SlipCell& moveRight(SlipIterator& X){ currentCell().moveRight(check(X)); return currentCell(); }


      SlipCell& moveListLeft(SlipCell& X)      { currentCell().moveListLeft(X); return currentCell(); }
      SlipCell& moveListRight(SlipCell& X)     { currentCell().moveListRight(X); return currentCell(); } // Move the current cell to the right of X
      SlipCell& moveListLeft(SlipIterator& X)  { currentCell().moveListLeft(X.currentCell()); return currentCell(); } // Move the current cell to the right of X
      SlipCell& moveListRight(SlipIterator& X) { currentCell().moveListRight(X.currentCell()); return currentCell(); } // Move the current cell to the right of X

      SlipCell& replace(const SlipIterator& X);
      SlipCell& replace(const SlipCell& X);    // Replace a cell  with a new cell
      SlipCell& replace(bool X)   { return (SlipDatum&)currentCell() = X; }
      SlipCell& replace(CHAR X)   { return (SlipDatum&)currentCell() = X; }
      SlipCell& replace(UCHAR X)  { return (SlipDatum&)currentCell() = X; }
      SlipCell& replace(LONG X)   { return (SlipDatum&)currentCell() = X; }
      SlipCell& replace(ULONG X)  { return (SlipDatum&)currentCell() = X; }
      SlipCell& replace(DOUBLE X) { return (SlipDatum&)currentCell() = X; }
      SlipCell& replace(const PTR X, const void* operation = ptrOP)   // Replace a cell  with a new cell
                                  { return (SlipDatum&)currentCell().replace(X, ptrOP); }
      SlipCell& replace(const string& X, bool constFlag = false)      // Replace a cell  with a new cell
                                  { return (SlipDatum&)currentCell().replace(X, constFlag); }
      SlipCell& replace(const string* X, bool constFlag = false)      // Replace a cell  with a new cell
                                  { return (SlipDatum&)currentCell().replace(*X, constFlag); }

      SlipCell& unlink() {
         SlipCell& cell = currentCell();
         advanceLWL();
         return cell.unLink();
      };

   /*****************************************
      *            Interrogatories            *
   ******************************************/
      bool      isData()       const { return currentCell().isData();     } // true if cell is data
      bool      isDiscrete()   const { return currentCell().isDiscrete(); } // true if cell contains a discrete number
      bool      isHeader()     const { return currentCell().isHeader();   } // true if cell is a header
      bool      isName()       const { return currentCell().isName();     } // true if cell is sublist
      bool      isNumber()     const { return currentCell().isNumber();   } // true if cell is numeric data
      bool      isPtr()        const { return currentCell().isPtr();      } // true if pointer
      bool      isReal()       const { return currentCell().isReal();     } // true if cell is a real number
      bool      isString()     const { return currentCell().isString();   } // true if cell is a string
      bool      isSublist()    const { return currentCell().isSublist();  } // true if cell is sublist

   /*****************************************
      *             Miscellaneous             *
   ******************************************/
      ClassType getClassType() const { return currentCell().getClassType(); }// get class type
      string    getName()      const { return currentCell().getName(); }  // return ascii name of cell
      SlipCell* getLeftLink()  const { return (SlipCell*)currentCell().getLeftLink(); } // Return pointer to previous cell
      SlipCell* getRightLink() const { return (SlipCell*)currentCell().getRightLink(); } // Return pointer to next cell
      SlipHeader& getHeader()  const {
         if (isSublist()) return ((SlipSublist&)currentCell()).getHeader();
         else postError(__FILE__, __LINE__, SlipErr::E1044, "(SlipSublist&)", "", "SlipCell not a sublist.", currentCell());
         return ((SlipSublist&)currentCell()).getHeader();
      }

   /*****************************************
   *           Virtual Functions           *
   ******************************************/

      virtual SlipCell& currentCell() const = 0;
      virtual string    toString()    const = 0;
      virtual string    dump()        const = 0;
      virtual SlipCell& deleteCell()        = 0;
}; // class SlipIteration : public SlipCellBase

}; // namespace slip
# endif /* SLIP_ITERATOR_H */
