 /*Copyright (C) 2023 Arthur I. Schwarz

   This file is part of the C++ SLIP library.  This library is free
   software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the
   Free Software Foundation; either version 3, or (at your option)
   any later version.
  
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
  
   Under Section 7 of GPL version 3, you are granted additional
   permissions described in the GCC Runtime Library Exception, version
   3.1, as published by the Free Software Foundation.
  
   You should have received a copy of the GNU General Public License and
   a copy of the GCC Runtime Library Exception along with this program;
   see the files COPYING3 and COPYING.RUNTIME respectively.  If not, see
   <http:www.gnu.org/licenses/>.
 */
 
 # include <string>
 # include "SlipParser.h"
 
 /**
  * @file SlipScanner.h
  * @author  A. Schwarz
  * @date    June 29, 2023
  * 
  * @brief This file provides function prototypes to support multi-buffer
  *        operations in handling include files.
  * 
  *   Flex v2.6.4 does not provide a means to include header file information
  *   in SlipFlex.h, further, SlipFlex.h contains #undef's which make it 
  *   unsuitable for inclusion in SlipFlex.cpp and inclusion in other files
  *   only insofar as they do not require the macros which are undefined. This
  *   means that function prototypes supporting multi-buffer operations with
  *   both the bison generated parser and flex generated scanner must provide
  *   another mechanism to suupport use of these prototypes, or to hand code
  *   them into each interation of SlipFlex.h. We have chosen to do our work
  *   in this header file.
  *   
  *   The functions are written using a C paradigm, hence they exist in global
  *   space, but are meant to be compiled by a C++ compiler. Effectively 
  *   namespaces are not used but classes are.
  *   
  *   The functions provide support for multi-buffer operations that require
  *   access to objects in both the bison and flex generated code. The functions
  *   are physically located in the flex generated code. Bison objects are
  *   passed as a function parameter.
  *   
  *   In operation bison and flex objects are treated together. A stack is
  *   used to support retention of data when an include file is seen. In
  *   operation, there are two stacks. One handled by flex generated functions,
  *   and one handled internally. Required push/pop operations are synched,
  *   a push/pop requires that both environments be saved or recovered.
  *   
  *   Creation of the class object used by the functions is done in the read
  *   function of SlipRead and then the parser is called. The creation of
  *   the object sets up the interface functions to allow the bison and flex
  *   generated code can support include files. A pointer to the object so
  *   created is maintained within the scope of the interface functions and
  *   allows only those functions access to the object.
  *   
  *   The functions and their description is:
  *   1. createState:   creates the object supporting I/F  operations
  *   2. saveState:     save the current state.
  *   3. restoreState:  restore a previous state.
  *   4. deleteState:   remove all object owned data.
  *   5. setLexorDebuON: turn on debugging in the scanner.
  *   6. setLexorDebuON: turn off debugging in the scanner.
  */  
  
  bool createState(string& filename);
  bool saveState(string& filename);
  bool restoreState();
  void deleteState();
  
  void setLexerDebugON();
  void setLexerDebugOFF();