// Copyright (C) 2014 Arthur I. Schwarz
//
// This file is part of the C++ SLIP library.  This library is free
// software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the
// Free Software Foundation; either version 3, or (at your option)
// any later version.

// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// Under Section 7 of GPL version 3, you are granted additional
// permissions described in the GCC Runtime Library Exception, version
// 3.1, as published by the Free Software Foundation.

// You should have received a copy of the GNU General Public License and
// a copy of the GCC Runtime Library Exception along with this program;
// see the files COPYING3 and COPYING.RUNTIME respectively.  If not, see
// <http://www.gnu.org/licenses/>.

/**
 * @file   SlipSequencer.cpp
 * @author A. Schwarz
 * @date   August 14, 2012
 *
 * @brief Implementation of the user interface to the sequencer iterator.
 */

# include <string>
# include <sstream>
# include <iomanip>
# include "SlipSequencer.h"
# include "SlipCell.h"
# include "SlipCellBase.h"
# include "SlipDatum.h"
# include "SlipGlobal.h"
# include "SlipHeader.h"
# include "SlipPointer.h"
# include "SlipSublist.h"
# include "SlipUndefinedOp.h"

namespace slip {

            /*****************************************
             *        Constructors/Destructors       *
            ******************************************/
   /**
    * @brief The <em>SlipSequencer</em> must created with a a <em>SlipHeader</em>.
    * <p>Construction of a <em>SlipSequencer</em> without a <em>SlipHeader</em> is
    *    not allowed. This constructor can not be used by the user.</p>
    */
   SlipSequencer::SlipSequencer() { };

   /**
    * @brief Constructor for a <em>SlipSequencer</em>.
    * <p>A <em>SlipSequencer</em> is constructing initally pointing to a
    *    list header.</p>
    * @param[in] header (SlipHeader&) initial cell in the sequencer
    */
   SlipSequencer::SlipSequencer(SlipHeader& header) {
      if (header.leftLink == UNDEFDATA)
         postError(__FILE__, __LINE__, SlipErr::E3026, "SlipSequncer::SlipSequencer", "", "", header);
      link = &header;
   }; // SlipSequencer::SlipSequencer(SlipHeader& header)

   /**
    * @brief Constructor for a <em>SlipSequencer</em>.
    * <p>A <em>SlipSequencer</em> is constructing initally pointing to a
    *    list header.</p>
    * @param[in] header (SlipHeader&) initial cell in the sequencer
    */
   SlipSequencer::SlipSequencer(const SlipHeader& header) {
      if (header.leftLink == UNDEFDATA)
         postError(__FILE__, __LINE__, SlipErr::E3026, "SlipSequncer::SlipSequencer", "", "", header);
      link = &const_cast<SlipHeader &>(header);
   }; // SlipSequencer::SlipSequencer(const SlipHeader& header)

   /**
    * @brief Constructor for a <em>SlipSequencer</em>.
    * <p>A <em>SlipSequencer</em> is constructing initally pointing to a
    *    list header extracted from the sublist cell.</p>
    * @param[in] sublist (SlipSublist&) reference to a header
    */
   SlipSequencer::SlipSequencer(SlipSublist& sublist) {
      link = (SlipCell*)*(sublist.getSublistHeader());
   }; // SlipSequencer::SlipSequencer(SlipSublist& sublist)

   /**
    * @brief Constructor for a <em>SlipSequencer</em>.
    * <p>A <em>SlipSequencer</em> is constructing initally pointing to a
    *    list header extracted from the sublist cell.</p>
    * @param[in] sublist (SlipSublist&) reference to a header
    */
   SlipSequencer::SlipSequencer(const SlipSublist& sublist) {
      link = (SlipCell*)*(sublist.getSublistHeader());
   }; // SlipSequencer::SlipSequencer(SlipSublist& sublist)

   /**
    * @brief Destructor
    * <p>A <em>SlipSequencer</em> has no internal body other than a pointer
    *    to a <em>SlipCell</em>. There isn't an associated structure to
    *    be deleted.</p>
    */
   SlipSequencer::~SlipSequencer() { }

            /*****************************************
             *          Sequencer Iteration          *
            ******************************************/
   /**
    * @brief Advance to the preceeding <em>SlipDatum</em> cell in the current list.
    * <p>Iterate over all cells in a list until a <em>SlipDatum</em> cell is
    *    found. Terminate on the list header (<em>SlipHeader</em>) if there
    *    are no more <em>SlipDatum</em> cells in the list.</p>
    * <p>Iteration starts on the cell preceeding the current cell.</p>
    * @return (SlipCell&) current cell after the advance
    */
   SlipCell& SlipSequencer::advanceLEL() {                                 // advance linear element left,  true if element found
      SlipCellBase* cell = (SlipCellBase*)link;
      for (cell = cell->getLeftLink(); !cell->isHeader(); cell = cell->getLeftLink())
         if (cell->isData()) break;
      link = (SlipCell*)cell;
      return currentCell();
   }; // SlipCell& SlipSequencer::advanceLEL()

   /**
    * @brief Advance to the next <em>SlipDatum</em> cell in the current list.
    * <p>Iterate over all cells in a list until a <em>SlipDatum</em> cell is
    *    found. Terminate on the list header (<em>SlipHeader</em>) if there
    *    are no more <em>SlipDatum</em> cells in the list.</p>
    * <p>Iteration starts on the cell following the current cell. </p>
    * @return (SlipCell&) current cell after the advance
    */
   SlipCell& SlipSequencer::advanceLER() {                                 // advance linear element right, true if element found
      SlipCellBase* cell = (SlipCellBase*)link;
      for (cell = cell->getRightLink(); !cell->isHeader(); cell = cell->getRightLink())
         if (cell->isData()) break;
      link = (SlipCell*)cell;
      return currentCell();
   }; // SlipCell& SlipSequencer::advanceLER()

   /**
    * @brief Advance to the preceeding <em>SlipSublist</em> cell in the current list.
    * <p>Iterate over all cells in a list until a <em>SlipSublist</em> cell is
    *    found. Terminate on the list header (<em>SlipHeader</em>) if there are
    *    no more <em>SlipSublist</em> cells in the list.</p>
    * <p>Iteration starts on the cell preceeding the current cell.</p>
    * @return (SlipCell&) current cell after the advance
    */
   SlipCell& SlipSequencer::advanceLNL() {                                 // advance linear name left,     true if element found
      SlipCellBase* cell = (SlipCellBase*)link;
      for (cell = cell->getLeftLink(); !cell->isHeader(); cell = cell->getLeftLink())
         if ((cell->isSublist())) break;
      link = (SlipCell*)cell;
      return currentCell();
   }; // SlipCell& SlipSequencer::advanceLNL()

   /**
    * @brief Advance to the next <em>SlipSublist</em> cell in the current list.
    * <p>Iterate over all cell in a list until a <em>SlipSublist</em> cell is
    *    found. Terminate on the list header (<em>SlipHeader</em>) if there
    *    are no more <em>SlipDatum</em> cells in the list.</p>
    * <p>Iteration starts on the cell following the current cell.</p>
    * @return (SlipCell&) current cell after the advance
    */
   SlipCell& SlipSequencer::advanceLNR() {                                 // advance linear name right,    true if element found
      SlipCellBase* cell = (SlipCellBase*)link;
      for (cell = cell->getRightLink(); !cell->isHeader(); cell = cell->getRightLink())
         if ((cell->isSublist())) break;
      link = (SlipCell*)cell;
      return currentCell();
   }; // SlipCell& SlipSequencer::advanceLNR()

   /**
    * @brief Advance to the preceeding cell in the current list.
    * <p>Iteration starts on the cell preceeding the current cell.</p>
    * @return (SlipCell&) current cell after the advance
    */
   SlipCell& SlipSequencer::advanceLWL() {                                 // advance linear word right,    true if element found
      SlipCellBase* cell = (SlipCellBase*)link;
      cell = cell->getLeftLink();
      link = (SlipCell*)cell;
      return currentCell();
   }; // SlipCell& SlipSequencer::advanceLWL()

   /**
    * @brief Advance to the following cell in the current list.
    * <p>Iteration starts on the cell following the current cell.</p>
    * @return (SlipCell&) current cell after the advance
    */
   SlipCell& SlipSequencer::advanceLWR() {                                 // advance linear word right,    true if element found
      SlipCellBase* cell = (SlipCellBase*)link;
      cell = cell->getRightLink();
      link = (SlipCell*)cell;
      return currentCell();
   }; // SlipCell& SlipSequencer::advanceLWR()

   /**
    * @brief Structurally advance to the preceeding <em>SlipDatum</em> cell.
    * <p>Iterate over all cells in a list until either a <em>SlipDatum</em>
    *    or a <em>SlipSublist</em> cell is found.
    * <ul>
    *    <li>If the current cell is a <em>SlipSublist</em> cell continue
    *        the search in the referenced sublist.</li>
    *    <li>If a <em>SlipDatum</em> cell is found, terminate the search.</li>
    *    <li>If a <em>SlipSublist</em> cell is found then continue the search
    *        in the referenced sublist.</li>
    *    <li>If a list header (<em>SlipHeader</em>) is found, stop.</li>
    * </ul>
    * @return (SlipCell&) current cell after the advance
    */
   SlipCell& SlipSequencer::advanceSEL() {                                 // advance structural element left,  true if element found
      SlipCellBase* cell = (SlipCellBase*)link;
      cell = cell->getLeftLink();
      while(!cell->isData()) {
         if (cell->isSublist()) {
            cell = (SlipCell*)*(SlipHeader**)SlipCellBase::getSublistHeader(*cell);
         } else if (cell->isHeader()) {
            break;
         }
         cell = cell->getLeftLink();
      }
      link = (SlipCell*)cell;
      return currentCell();
   }; // SlipCell&& SlipSequencer::advanceSEL()

   /**
    * @brief Structurally advance to the following <em>SlipDatum</em> cell.
    * <p>Iterate over all cells in a list until either a <em>SlipDatum</em>
    *    or a <em>SlipSublist</em> cell is found.
    * <ul>
    *    <li>If the current cell is a <em>SlipSublist</em> cell continue
    *        the search in the referenced sublist.</li>
    *    <li>If a <em>SlipDatum</em> cell is found, terminate the search.</li>
    *    <li>If a <em>SlipSublist</em> cell is found then continue the search
    *        in the referenced sublist.</li>
    *    <li>If a list header (<em>SlipHeader</em>) is found, stop.</li>
    * </ul>
    * @return (SlipCell&) current cell after the advance
    */
   SlipCell& SlipSequencer::advanceSER() {                                 // advance structural element right, true if element found
      SlipCellBase* cell = (SlipCellBase*)link;
      cell = cell->getRightLink();
      while(!cell->isData()) {
         if (cell->isSublist()) {
            cell = (SlipCell*)*(SlipHeader**)SlipCellBase::getSublistHeader(*cell);
         } else if (cell->isHeader()) {
           break;
         }
         cell = cell->getRightLink();
      }
      link = (SlipCell*)cell;
      return currentCell();
   }; // SlipCell& SlipSequencer::advanceSER()

   /**
    * @brief Structurally advance to the preceeding <em>SlipSublist</em> cell.
    * <p>Iterate over all cells in a list until a <em>SlipSublist</em> cell is
    *    found. </p>
    * <ul>
    *    <li>If the current cell is a <em>SlipSublist</em> cell continue
    *        the search in the referenced sublist.</li>
    *    <li>If a <em>SlipSublist</em> cell is found, terminate the search.</li>
    *    <li>If a list header (<em>SlipHeader</em>), stop.</li>
    * </ul>
    * <p>Iteration starts on the current cell.</p>
    * @return (SlipCell&) current cell after the advance
    */
   SlipCell& SlipSequencer::advanceSNL() {                                 // advance structural name left,     true if element found
      SlipCellBase* cell = (SlipCellBase*)link;
      if (cell->isSublist())
         cell = (SlipCell*)*(SlipHeader**)SlipCellBase::getSublistHeader(*cell);
      if (cell->isHeader())
         cell   = cell->getLeftLink();
      while(!cell->isHeader()) {
         if (cell->isSublist())
            break;
         cell = cell->getLeftLink();
      }
      link = (SlipCell*)cell;
      return currentCell();
   }; // SlipCell& SlipSequencer::advanceSNL()

   /**
    * @brief Structurally advance to the following <em>SlipSublist</em> cell.
    * <p>Iterate over all cells in a list until a <em>SlipSublist</em> cell is
    *    found. </p>
    * <ul>
    *    <li>If the current cell is a <em>SlipSublist</em> cell continue
    *        the search in the referenced sublist.</li>
    *    <li>If a <em>SlipSublist</em> cell is found, terminate the search.</li>
    *    <li>If a list header (<em>SlipHeader</em>), stop.</li>
    * </ul>
    * <p>Iteration starts on the current cell.</p>
    * @return (SlipCell&) current cell after the advance
    */
   SlipCell& SlipSequencer::advanceSNR() {                                 // advance structural name right,    true if element found
      SlipCellBase* cell = (SlipCellBase*)link;
      if (cell->isSublist())
         cell = (SlipCell*)*(SlipHeader**)SlipCellBase::getSublistHeader(*cell);
      if (cell->isHeader())
         cell   = cell->getRightLink();
      while(!cell->isHeader()) {
         if (cell->isSublist())
            break;
         cell = cell->getRightLink();
      }
      link = (SlipCell*)cell;
      return currentCell();
   }; // SlipCell&& SlipSequencer::advanceSNR()

   /**
    * @brief Structurally advance to the previous cell.
    * <p>Iteration will advance only one cell. </p>
    * <ul>
    *    <li>If the current cell is a sublist (<em>SlipSublist</em>) then
    *        iteration will follow the nested list header.</li>
    *    <li>If the current cell is a list header (<em>SlipHeader</em>), stop.</li>
    *    <li>Otherwise iteration will stop at the preceeding cell.</li>
    * </ul>
    * @return (SlipCell&) current cell after the advance
    */
   SlipCell& SlipSequencer::advanceSWL() {                                 // advance structural word right,    true if element found
      SlipCellBase* cell = (SlipCellBase*)link;
      if (cell->isSublist())
         cell = (SlipCell*)*(SlipHeader**)SlipCellBase::getSublistHeader(*cell);
      cell = cell->getLeftLink();
      link = (SlipCell*)cell;
      return currentCell();
   }; // SlipCell& SlipSequencer::advanceSWL()

   /**
    * @brief Structurally advance to the following cell.
    * <p>Iteration will advance only one cell. </p>
    * <ul>
    *    <li>If the current cell is a sublist (<em>SlipSublist</em>) then
    *        iteration will follow the nested list header.</li>
    *    <li>If the current cell is a list header (<em>SlipHeader</em>), stop.</li>
    *    <li>Otherwise iteration will stop at the following cell.</li>
    * </ul>
    * @return (SlipCell&) current cell after the advance
    */
   SlipCell& SlipSequencer::advanceSWR() {                                 // advance structural word right,    true if element found
      SlipCellBase* cell = (SlipCellBase*)link;
      if (cell->isSublist())
         cell = (SlipCell*)*(SlipHeader**)SlipCellBase::getSublistHeader(*cell);
      cell = cell->getRightLink();
      link = (SlipCell*)cell;
      return currentCell();
   }; // SlipCell& SlipSequencer::advanceSWR()

    /**
     * @fn SlipSequencer::deleteCell()
     * @brief Delete the current cell.
     * <p>The current cell is deleted and the sequencer current cell points to
     *    the previous cell.</p>
     *
     * @return (SlipCell&) reference to the current cell.
     */
SlipCell& SlipSequencer::deleteCell() {                          // delete the current cell
   SlipCell* cell = &currentCell();;
   advanceLWL();
   delete cell;
   return currentCell();
}; // SlipCell& SlipSequencer::deleteCell()

}; // namespace slip
