// Copyright (C) 2014 Arthur I. Schwarz
//
// This file is part of the C++ SLIP library.  This library is free
// software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the
// Free Software Foundation; either version 3, or (at your option)
// any later version.

// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// Under Section 7 of GPL version 3, you are granted additional
// permissions described in the GCC Runtime Library Exception, version
// 3.1, as published by the Free Software Foundation.

// You should have received a copy of the GNU General Public License and
// a copy of the GCC Runtime Library Exception along with this program;
// see the files COPYING3 and COPYING.RUNTIME respectively.  If not, see
// <http://www.gnu.org/licenses/>.

/**
 * @file   SlipSequencer.h
 * @author A. Schwarz
 * @date   August 14, 2012
 *
 * @brief User interface to the non-structured list iterator.
 *
 */

#ifndef SLIPSEQUENCER_H
#define	SLIPSEQUENCER_H

# include "SlipHeader.h"
# include "SlipCell.h"
# include "SlipCellBase.h"
# include "SlipDatum.h"
# include "SlipGlobal.h"
# include "SlipIterator.h"
# include "SlipSublist.h"

namespace slip {
/**
 * @class slip::SlipSequencer
 * 
 * @brief User interface to the non-structured list iterator.
 * 
 * <p><b>Overview</b></p><hr>
 * <p>Linear iteration is the process of visiting cells in a single list.
 *    Structured iteration is the processes of visiting cells in a list
 *    and sublist(s).The sequencer interface provides a means to do
 *    linear and structured iteration over a list.</p>
 * <p>The sequencer interface provides linear and structured left and right 
 *    (previous and next) iteration over an acyclic graph with the means
 *    to determine the cell type to terminate the iteration. Iteration
 *    is called <b>advance</b> and there are three attributes that can 
 *    be used in an iteration:</p>
 * <ul>
 *    <li><b>Element (E)</b> advance until a <em>SlipDatum</em> cell is found.</li>
 *    <li><b>Name (N)</b> advance until a <em>SlipSublist</em> cell is found.</li>
 *    <li><b>Word (*W)</b> advance to the next cell.</li>
 * </ul>
 * <p>If the search element (E or W) is not found, then the iteration will stop
 *    at the current list header (SlipHeader) for linear (L) iterations
 *    and at the bottommost list header for structured (S) iterations.</p>
 * <p>The advance can be <b>Left (L)</b> or <b>Right (R)</b>, or 
 *    <b>Linear (L)</b> or <b>Structured (S)</b>. This can be composed
 *    into the following advance methods:</p>
 * <tt><center>
 * |   advance  |  iterate  |  search | direction |
 * | :--------- | :-------- | :------ | :-------- |
 * | advanceLER | linear    | element | right     |
 * | advanceLEL | linear    | element | left      |
 * | advanceLNR | linear    | name    | right     |
 * | advanceLNL | linear    | name    | left      |
 * | advanceLWR | linear    | word    | right     |
 * | advanceLWL | linear    | word    | left      |
 * | advanceSER | structure | element | right     |
 * | advanceSEL | structure | element | left      |
 * | advanceSNR | structure | name    | right     |
 * | advanceSNL | structure | name    | left      |
 * | advanceSWR | structure | word    | right     |
 * | advanceSWL | structure | word    | left      |
 * </center></tt>
 * <p>The legends have the following meaning:</p>
 * <ul type="none">
 *    <li><b>advance</b> name of the method.</li>
 *    <li><b>iteration</b> linear iteration will not ascend from or descend
 *        to a list, structured iteration will ascend from or descend to a
 *        list.</li>
 *    <li><b>search</b> stop iterating when an element, name, or word is
 *        found. Iteration will always terminate on a list head, 
 *        <b>SlipHeader</b>, cell if the search item is not found. For
 *        linear iterators a search will stop when the list head of the 
 *        current list is found. For structured iterators a search will
 *        stop at the bottommost list head on the first sublist seen in
 *        each indentured list.</li>
 *    <li><b>direction</b> the direction of the iteration starting at the
 *        current cell. The iteration will proceed to prior cells for 
 *        <b>Left</b> and to succeeding next cells for <b>Right</b>.
 *        Sublist entry/exit depends on the iteration type, <b>Linear</b> 
 *        <b>Structured</b>.</li>
 * </ul>
 * <p>An example of the use of these functions is given:</p>
 * <tt><pre>
 *   <b>list = (a b (c (d) e ) f)</b>
 * </pre></tt>
 * <p>Using the notation D(value) to represent a SlipDatum value, and H() to
 *    represent a SlipHeader defined list, and N(H()) to represent a SlipSublist
 *    reference, N(), to a list header, H() where H represents the 'name' of
 *    a list, we transform the above list to the following linear relation:</p>
 * <tt><pre>
 *   <b>list = L1( D(a) D(b) N(L2(D(c) N(L3(D(d))) D(e)) D(f))</b>
 * </pre></tt>
 * <p>The spatial relation representing this list is:</p>
 * <tt><pre>
 *   <b>L1( D(a) D(b) N() D(f) )word
 *                 |
 *                 |
 *                 o L2( D(c) N() D(e) )
 *                            |
 *                            |
 *                            o L3( D(d) )
 * </b>
 * </pre></tt>
 * <p>The following examples show the effect of each of the readers over the
 *    list. The sublist references will be returned as N(H) representing
 *    a sublist, N, and the list it references, H. In our example there are
 *    two sublists, N(L2) and N(L3) representing the sublist cells referencing
 *    lists L2 and L3 respectively.</p>
 * <p>All sequencers start at L1.</p>
 * <ul>
 *    <li><b>SlipSequencer R = new SlipSequencer(L1);</b> &rarr; <tt>&lt;L1 L1&gt;</tt></li>
 *    <ul type="none">
 *       <li><b>R.advanceLER();</b> &rarr; <tt>&lt;D(a)&nbsp;&nbsp;&gt;</tt></li>
 *       <li><b>R.advanceLER();</b> &rarr; <tt>&lt;D(b)&nbsp;&nbsp;&gt;</tt></li>
 *       <li><b>R.advanceLER();</b> &rarr; <tt>&lt;D(f)&nbsp;&nbsp;&gt;</tt></li>
 *       <li><b>R.advanceLER();</b> &rarr; <tt>&lt;L1&nbsp;&nbsp;&nbsp;&nbsp;&gt;</tt></li>
 *       <li><b>delete R;</b></li>
 *    </ul>
 *    <li><b>SlipSequencer R = new SlipSequencer(L1);</b> &rarr; <tt>&lt;L1&gt;</tt></li>
 *    <ul type="none">
 *       <li><b>R.advanceLEL();</b> &rarr; <tt>&lt;D(f)&nbsp;&nbsp;&gt;</tt></li>
 *       <li><b>R.advanceLEL();</b> &rarr; <tt>&lt;D(b)&nbsp;&nbsp;&gt;</tt></li>
 *       <li><b>R.advanceLEL();</b> &rarr; <tt>&lt;D(a)&nbsp;&nbsp;&gt;</tt></li>
 *       <li><b>R.advanceLEL();</b> &rarr; <tt>&lt;L1&nbsp;&nbsp;&nbsp;&nbsp;&gt;</tt></li>
 *       <li><b>delete R;</b></li>
 *    </ul>
 *    <li><b>SlipSequencer R = new SlipSequencer(L1);</b> &rarr; <tt>&lt;L1&gt;</tt></li>
 *    <ul type="none">
 *       <li><b>R.advanceLNR();</b> &rarr; <tt>&lt;N(L2)&gt;</tt></li>
 *       <li><b>R.advanceLNR();</b> &rarr; <tt>&lt;L1&nbsp;&nbsp;&nbsp;&gt;</tt></li>
 *       <li><b>delete R;</b></li>
 *    </ul>
 *    <li><b>SlipSequencer R = new SlipSequencer(L1);</b> &rarr; <tt>&lt;L1&gt;</tt></li>
 *    <ul type="none">
 *       <li><b>R.advanceLNL();</b> &rarr; <tt>&lt;N(L2)&gt;</tt></li>
 *       <li><b>R.advanceLNL();</b> &rarr; <tt>&lt;L1&nbsp;&nbsp;&nbsp;&gt;</tt></li>
 *       <li><b>delete R;</b></li>
 *    </ul>
 *    <li><b>SlipSequencer R = new SlipSequencer(L1);</b> &rarr; <tt>&lt;L1&nbsp;&nbsp;&nbsp;&gt;</tt></li>
 *    <ul type="none">
 *       <li><b>R.advanceLWR();</b> &rarr; <tt>&lt;D(a)&nbsp;&nbsp;&gt;</tt></li>
 *       <li><b>R.advanceLWR();</b> &rarr; <tt>&lt;D(b)&nbsp;&nbsp;&gt;</tt></li>
 *       <li><b>R.advanceLWR();</b> &rarr; <tt>&lt;N(L2)&nbsp;&gt;</tt></li>
 *       <li><b>R.advanceLWR();</b> &rarr; <tt>&lt;D(f)&nbsp;&nbsp;&gt;</tt></li>
 *       <li><b>R.advanceLWR();</b> &rarr; <tt>&lt;L1&nbsp;&nbsp;&nbsp;&nbsp;&gt;</tt></li>
 *       <li><b>delete R;</b></li>
 *    </ul>
 *    <li><b>SlipSequencer R = new SlipSequencer(L1);</b> &rarr; <tt>&lt;L1&gt;</tt></li>
 *    <ul type="none">
 *       <li><b>R.advanceLWL();</b> &rarr; <tt>&lt;D(f)&nbsp;&nbsp;&gt;</tt></li>
 *       <li><b>R.advanceLWL();</b> &rarr; <tt>&lt;N(L2)&nbsp;&gt;</tt></li>
 *       <li><b>R.advanceLWL();</b> &rarr; <tt>&lt;D(b)&nbsp;&nbsp;&gt;</tt></li>
 *       <li><b>R.advanceLWL();</b> &rarr; <tt>&lt;D(a)&nbsp;&nbsp;&gt;</tt></li>
 *       <li><b>R.advanceLWL();</b> &rarr; <tt>&lt;L1&nbsp;&nbsp;&nbsp;&nbsp;&gt;</tt></li>
 *       <li><b>delete R;</b></li>
 *    </ul>
 *    <li><b>SlipSequencer R = new SlipSequencer(L1);</b> &rarr; <tt>&lt;L1&gt;</tt></li>
 *    <ul type="none">
 *       <li><b>R.advanceSER();</b> &rarr; <tt>&lt;D(a)&nbsp;&nbsp;&gt;</tt></li>
 *       <li><b>R.advanceSER();</b> &rarr; <tt>&lt;D(b)&nbsp;&nbsp;&gt;</tt></li>
 *       <li><b>R.advanceSER();</b> &rarr; <tt>&lt;D(c)&nbsp;&nbsp;&gt;</tt></li>
 *       <li><b>R.advanceSER();</b> &rarr; <tt>&lt;D(d)&nbsp;&nbsp;&gt;</tt></li>
 *       <li><b>R.advanceSER();</b> &rarr; <tt>&lt;L1&nbsp;&nbsp;&nbsp;&nbsp;&gt;</tt></li>
 *       <li><b>delete R;</b></li>
 *    </ul>
 *    <li><b>SlipSequencer R = new SlipSequencer(L1);</b> &rarr; <tt>&lt;L1&gt;</tt></li>
 *    <ul type="none">
 *       <li><b>R.advanceSEL();</b> &rarr; <tt>&lt;D(f)&nbsp;&nbsp;&gt;</tt></li>
 *       <li><b>R.advanceSEL();</b> &rarr; <tt>&lt;D(e)&nbsp;&nbsp;&gt;</tt></li>
 *       <li><b>R.advanceSEL();</b> &rarr; <tt>&lt;D(d)&nbsp;&nbsp;&gt;</tt></li>
 *       <li><b>R.advanceSEL();</b> &rarr; <tt>&lt;L3&nbsp;&nbsp;&nbsp;&nbsp;&gt;</tt></li>
 *       <li><b>delete R;</b></li>
 *    </ul>
 *    <li><b>SlipSequencer R = new SlipSequencer(L1);</b> &rarr; <tt>&lt;L1&nbsp;&nbsp;&nbsp;&gt;</tt></li>
 *    <ul type="none">
 *       <li><b>R.advanceSNR();</b> &rarr; <tt>&lt;N(L2)&gt;</tt></li>
 *       <li><b>R.advanceSNR();</b> &rarr; <tt>&lt;N(L3)&gt;</tt></li>
 *       <li><b>R.advanceSNR();</b> &rarr; <tt>&lt;L2&nbsp;&nbsp;&nbsp;&gt;</tt></li>
 *       <li><b>delete R;</b></li>
 *    </ul>
 *    <li><b>SlipSequencer R = new SlipSequencer(L1);</b> &rarr; <tt>&lt;L1&gt;</tt></li>
 *    <ul type="none">
 *       <li><b>R.advanceSNL();</b> &rarr; <tt>&lt;N(L2)&gt;</tt></li>
 *       <li><b>R.advanceSNL();</b> &rarr; <tt>&lt;N(L3)&gt;</tt></li>
 *       <li><b>R.advanceSNL();</b> &rarr; <tt>&lt;L2&nbsp;&nbsp;&nbsp;&gt;</tt></li>
 *       <li><b>delete R;</b></li>
 *    </ul>
 *    <li><b>SlipSequencer R = new SlipSequencer(L1);</b> &rarr; <tt>&lt;L1&gt;</tt></li>
 *    <ul type="none">
 *       <li><b>R.advanceSWR();</b> &rarr; <tt>&lt;D(a)&nbsp;&nbsp;&gt;</tt></li>
 *       <li><b>R.advanceSWR();</b> &rarr; <tt>&lt;D(b)&nbsp;&nbsp;&gt;</tt></li>
 *       <li><b>R.advanceSWR();</b> &rarr; <tt>&lt;N(L2)&nbsp;&gt;</tt></li>
 *       <li><b>R.advanceSWR();</b> &rarr; <tt>&lt;D(c)&nbsp;&nbsp;&gt;</tt></li>
 *       <li><b>R.advanceSWR();</b> &rarr; <tt>&lt;N(L3)&nbsp;&gt;</tt></li>
 *       <li><b>R.advanceSWR();</b> &rarr; <tt>&lt;D(d)&nbsp;&nbsp;&gt;</tt></li>
 *       <li><b>R.advanceSWR();</b> &rarr; <tt>&lt;L3&nbsp;&nbsp;&nbsp;&nbsp;&gt;</tt></li>
 *       <li><b>delete R;</b></li>
 *    </ul>
 *    <li><b>SlipSequencer R = new SlipSequencer(L1);</b> &rarr; <tt>&lt;L1&gt;</tt></li>
 *    <ul type="none">
 *       <li><b>R.advanceSWL();</b> &rarr; <tt>&lt;D(f)&nbsp;&nbsp;&gt;</tt></li>
 *       <li><b>R.advanceSWL();</b> &rarr; <tt>&lt;N(L2)&nbsp;&gt;</tt></li>
 *       <li><b>R.advanceSWL();</b> &rarr; <tt>&lt;D(e)&nbsp;&nbsp;&gt;</tt></li>
 *       <li><b>R.advanceSWL();</b> &rarr; <tt>&lt;N(L3)&nbsp;&gt;</tt></li>
 *       <li><b>R.advanceSWL();</b> &rarr; <tt>&lt;D(d)&nbsp;&nbsp;&gt;</tt></li>
 *       <li><b>R.advanceSWL();</b> &rarr; <tt>&lt;L3&nbsp;&nbsp;&nbsp;&nbsp;&gt;</tt></li>
 *       <li><b>delete R;</b></li>
 *    </ul>
 *    <li><b>SlipSequencer R = new SlipSequencer(L1);</b> &rarr; <tt>&lt;L1&gt;</tt></li>
 *    <ul type="none">
 *       <li><b>R.advanceSNR();</b> &rarr; <tt>&lt;N(L2)&gt;</tt></li>
 *       <li><b>R.advanceSER();</b> &rarr; <tt>&lt;D(c)&nbsp;&gt;</tt></li>
 *       <li><b>R.advanceLWL();</b> &rarr; <tt>&lt;L2&nbsp;&nbsp;&nbsp;&gt;</tt></li>
 *       <li><b>R.advanceLNR();</b> &rarr; <tt>&lt;N(L3)&gt;</tt></li>
 *       <li><b>R.advanceLNR();</b> &rarr; <tt>&lt;L2&nbsp;&nbsp;&nbsp;&gt;</tt></li>
 *       <li><b>delete R;</b></li>
 *    </ul>
 * </ul>
 * <p>Where &lt;&gt; represents &lt;current cell&gt;.</p>
 * <p>There are several additional methods which support sequencer operations.
 *    The user can reset the current cell reference, <em>reset()</em> and 
 *    retrieve a reference to the current cell, <em>currentCell()</em>. There 
 *    are other operations and methods which are conveniences to support the 
 *    user but which do not contribute to sequencer functionality. These are 
 *    all documented in the following sections.</p>
 * 
 * <p><b>Constructors/Destructors</b></p><hr>
 * <p>A null constructor is not available to the user. The user must
 *    provide a list header directly or indirectly through a reference
 *    to a sublist cell to construct the sequencer. If construction is done 
 *    on the stack, the system will automatically delete the sequencer 
 *    on scope exit, otherwise the user must explicitely delete the sequencer./p>
 * <ul>
 *    <li><b>SlipSequencer()</b> unavailable to the user.</li>
 *    <li><b>SlipSequencer(header)</b> construct an instance of a sequencer.</li>
 *    <li><b>SlipSequencer(sublist)</b> construct an instance of the sequencer
 *        iterator using (<em>SlipSublist</em>&) sublist.</li>
 *    <li><b>~SlipSequencer()</b> delete the Sequencer.</li>
 * </ul>
 * 
 * <p><b>Iteration</b></p><hr>
 * <p>The linear and structural iteration methods have been described above.
 *    They are listed here:</p>
 * <ul>
 *    <li><b>advanceLER()</b> advance linearly to the right and stop at a 
 *        <em>SlipDatum</em> cell.</li>
 *    <li><b>advanceLEL()</b> advance linearly to the left and stop at a 
 *        <em>SlipDatum</em> cell.</li>
 *    <li><b>advanceLNR()</b> advance linearly to the right and stop at a 
 *        <em>SlipSublist</em> cell.</li>
 *    <li><b>advanceLNL()</b> advance linearly to the left and stop at a 
 *        <em>SlipSublist</em> cell.</li>
 *    <li><b>advanceLWR()</b> advance linearly to the right and stop at a 
 *        <em>SlipDatum</em>, <em>SlipSublist</em>, or <em>SlipHeader</em> 
 *        cell.</li>
 *    <li><b>advanceLWL()</b> advance linearly to the left and stop at a 
 *        <em>SlipDatum</em>, <em>SlipSublist</em>, or <em>SlipHeader</em> 
 *        cell.</li>
 *    <li><b>advanceSER()</b> advance structurally to the right and stop at a 
 *        <em>SlipDatum</em> cell.</li>
 *    <li><b>advanceSEL()</b> advance structurally to the left and stop at a 
 *        <em>SlipDatum</em> cell.</li>
 *    <li><b>advanceSNR()</b> advance structurally to the right and stop at a 
 *        <em>SlipSublist</em> cell.</li>
 *    <li><b>advanceSNL()</b> advance structurally to the left and stop at a 
 *        <em>SlipSublist</em> cell.</li>
 *    <li><b>advanceSWR()</b> advance structurally to the right and stop at a 
 *        <em>SlipDatum</em>, <em>SlipSublist</em>, or <em>SlipHeader</em> 
 *        cell.</li>
 *    <li><b>advanceSWL()</b> advance structurally to the left and stop at a 
 *        <em>SlipDatum</em>, <em>SlipSublist</em>, or <em>SlipHeader</em> 
 *        cell.</li>
 * </ul>
 * 
 * <p><b>Iteration Miscellaneous</b></p>
 * <p>Several methods support the use of the sequencer. They do not iterate over
 *    a list but support the user.</p>
 * <ul>
 *    <li><b>currentCell()</b> return a reference to the current sequencer 
 *        current cell.</li>
 *    <li><b>reset()</b> reset the current cell to a <b>SlipCell</b>.</li>
 * </ul>
 * 
 * <p><b>Convenience Methods</b></p>
 * <p>These methods provide increased usability. They provide access to
 *    existing methods without requiring that the user extract a cell 
 *    reference (R.currentCell()), and they maintain sequencer consistency 
 *    when the operation invalidates the <em>currentCell()</em>.</p>
 * <p>Consistency assures the user that the operation and the sequencer are
 *    in synchrony. Operations done outside of the sequencer may place the
 *    sequencer in an inconsistent state. Operations done with the sequencer
 *    ensure that the sequencer is in a consistent state before and after
 *    the operation.</p>
 * 
 * <p><b>Cell Convenience Methods</b></p><hr>
 * <p>Cell methods maintain the consistency of the sequencer before and after
 *    execution. The methods are a short-hand version of SLIP operations using 
 *    the current cell as a default in the operation.</p>
 * <ul>
 *    <li><b>deleteCell()</b> delete the current cell and perform a 
 *        <em>advanceLWL()</em>.</li>
 *    <li><b>insLeft()</b> insert the input cell to the left of the 
 *        current cell.</li>
 *    <li><b>insRight()</b> insert the input cell to the right of the 
 *        current cell.</li></li>
 *    <li><b>moveLeft()</b> move the input cell to the left of the
 *        current cell.</li>
 *    <li><b>moveRight()</b> move the input cell to the right of the
 *        current cell.</li>
 *    <li><b>replace()</b> replace the current cell with the input value
 *        and set the current cell to the replacement cell.</li>
 *    <li><b>unlink()</b> unlinks the current cell from a list and advances
 *        the sequencer one word to the left.</li>
 * </ul>
 * 
 * <p><b>Operators</b></p><hr>
 * <p></p>
 * <ul>
 *    <li><b>operator=()</b> Assignment can be to the sequencer or to a cell.</li>
 *    <ul>
 *        <li>List assignment replaces the sequencer current cell with a 
 *            pointer to the SlipHeader object.</li>
 *        <li>Sequencer assignment replaces the current cell with a the RHS current
 *            cell reference.</li>
 *        <li>Sequencer assignment replaces the current cell with a the RHS current
 *            cell reference.</li>
 *        <li>Cell assignment assigns like-to-like. A RHS <em>SlipDatum</em> 
 *            cell can be assigned to a LHS <em>SlipDatum</em> cell and a RHS 
 *            <em>SlipSublist</em> cell can be assigned to a LHS 
 *            <em>SlipSublist</em> cell. Unlike ordinary cell assignment
 *            (see slip::SlipDatum) the assignment of a <em>SlipHeader</em>
 *            reference to a <em>SlipSublist</em> cell can not be done.</li>
 *    </ul>
 *    <li><b>operator==()</b> compares the current cell to the RHS. 
 *        If a <em>SlipSequencer</em> is given as the RHS then the comparison is 
 *        <b>true</b> if both the current cells are equal.></li>
 *    <li><b>operator!=()</b> compares the LHS to the RHS <em>slip::SlipCell</em>. 
 *        If a <em>SlipSequencer</em> is given as the RHS then the comparison is 
 *        <b>true</b> if both the current cells are unequal.></li>
 * </ul>
 * 
 * <p><b>Miscellaneous</b></p><hr>
 * <p></p>
 * <ul>
 *    <li><b>getClassType()</b> return the class type of the current cell.</li>
 *    <li><b>getName()</b> return a string representing the <em>SlipDatum</em>
 *        type of the current cell.</li>
 *    <li><b>isData()</b> <b>true</b> if the current cell is data.</li>
 *    <li><b>isHeader()</b> <b>true</b> if the current cell is a 
 *        <em>SlipHeader</em>/li>
 *    <li><b>isName()</b> <b>true</b> if the current cell is a 
 *        <em>SlipSublist</em>.</li>
 *    <li><b>isNumber()</b> <b>true</b> if the current cell is a 
 *        <em>SlipDatum</em> number.</li>
 *    <li><b>isPtr()</b> <b>true</b> if the current cell is a 
 *        <em>SlipDatum</em> pointer.</li>
 *    <li><b>isSublist()</b> <b>true</b> if the current cell is a 
 *        <em>SlipSublist</em>.</li>
 *    <li><b>toString()</b> pretty-print the current cell.</li>
 *    <li><b>dump()</b> uglify the current cell.</li>
 * </ul>
 * <p><b>NOTES</b></p><hr>
 * <ol>
 *    <li>Deletion of a list containing a cell referenced by a sequencer
 *        will cause the sequencer to go out of synchrony when the list
 *        cells are returned to the AVSL.</li>
 *    <li>Deletion of a cell referenced by a sequencer will cause the
 *        sequencer to go out of synchrony.
 *    </li>
 * </ol>
 * 
 */

 /*****************************************
 *            Cell Assignment            *
 ******************************************/
 /**
 * @fn SlipIteration::operator=(const SlipDatum& X)
 * @brief Assign (<b>Y = X</b>) where Y is the current cell.
 * <p>For purposes of this section, <b>Y = X</b> refers to the left hand
 *    side (LHS) being replaced by the right hand side (RHS) of the
 *    assignment symbol (<b>=</b>). The LHS refers to the current object
 *    and the RHS refers to the input object.</p>
 * <p>Assignments are class invariant. Assignments do not preserve type.
 *    Under assignment the LHS <em>datum</em> field will be
 *    changed according to the RHS <em>datum</em> field and the LHS type
 *    field will be changed to correspond to the RHS.</p>
 * <p>When the assignment is complete, the RHS will not have changed and
 *    the LHS type and value will be the same as the RHS. LHS == RHS will
 *    be <b>true</b>.
 * <p>If the LHS is a SlipPointer object (PTR or STRING) then the object
 *    pointed to must be deleted (for PTR if it is a user defined object and
 *    for STRING if it is a string). This decision is deferred to the
 *    SlipPointer <em>remove()</em> method. If the object is treated as a
 *    constant in SLIP, then <em>remove</em> does nothing. If the object is
 *    not a constant, the <em>remove()</em> deletes it. <em>remove()</em> is a
 *    proxy for <em>delete (void*)</em> within the context of SlipPointer.<p>
 * <p>If the RHS is a SlipPointer object then the object pointed to must
 *    be copied (for PTR it is a user defined object and for STRING it is
 *    a string). If the SlipPointer object is constant within SLIP, the
 *    the copy copies the pointer to the SlipPointer object. If the
 *    SlipPointer object is not a constant within SLIP, then the copy
 *    operation creates a new SlipPointer object. The result of the copy
 *    is inserted into the LHS <em>datum</em> field. This decision is
 *    deferred to the SlipPointer <em>copy()</em> method.</p>
 * @param[in] X (SlipDatum&) RHS of the assignment
 * @return reference to the current <em>SlipSequencer</em>
  */

/**
 * @fn SlipIteration::operator=(const SlipSequencer& X)
 * @brief Assign (<b>currentCell() = X.currentCell()</b>).
 * <p>For purposes of this section, <b>Y = X</b> refers to the left hand
 *    side (LHS) being replaced by the right hand side (RHS) of the
 *    assignment symbol (<b>=</b>). The LHS refers to the current object
 *    and the RHS refers to the input object.</p>
 * <p>Assignments are class invariant. Assignments do not preserve type.
 *    Under assignment the LHS <em>datum</em> field will be
 *    changed according to the RHS <em>datum</em> field and the LHS type
 *    field will be changed to correspond to the RHS.</p>
 * <p>When the assignment is complete, the RHS will not have changed and
 *    the LHS type and value will be the same as the RHS. LHS == RHS will
 *    be <b>true</b>.
 * <p>If the LHS is a SlipPointer object (PTR or STRING) then the object
 *    pointed to must be deleted (for PTR if it is a user defined object and
 *    for STRING if it is a string). This decision is deferred to the
 *    SlipPointer <em>remove()</em> method. If the object is treated as a
 *    constant in SLIP, then <em>remove</em> does nothing. If the object is
 *    not a constant, the <em>remove()</em> deletes it. <em>remove()</em> is a
 *    proxy for <em>delete (void*)</em> within the context of SlipPointer.<p>
 * <p>If the RHS is a SlipPointer object then the object pointed to must
 *    be copied (for PTR it is a user defined object and for STRING it is
 *    a string). If the SlipPointer object is constant within SLIP, the
 *    the copy copies the pointer to the SlipPointer object. If the
 *    SlipPointer object is not a constant within SLIP, then the copy
 *    operation creates a new SlipPointer object. The result of the copy
 *    is inserted into the LHS <em>datum</em> field. This decision is
 *    deferred to the SlipPointer <em>copy()</em> method.</p>
 * @param[in] X (SlipSequencer&) RHS of assignment
 * (SlipSequencer&) reference to the sequencer
 */

/**
 * @fn SlipIteration::operator=(const bool    X)
 * @brief Assign (<b>Y = X</b>) where Y is the current cell.
 * <p>For purposes of this section, <b>Y = X</b> refers to the left hand
 *    side (LHS) being replaced by the right hand side (RHS) of the
 *    assignment symbol (<b>=</b>). The LHS refers to the current object
 *    and the RHS refers to the input object.</p>
 * <p>Assignments are class invariant. Assignments do not preserve type.
 *    Under assignment the LHS <em>datum</em> field will be
 *    changed according to the RHS <em>datum</em> field and the LHS type
 *    field will be changed to correspond to the RHS.</p>
 * <p>When the assignment is complete, the RHS will not have changed and
 *    the LHS type will be bool and the value will be the same as the RHS.
 *    LHS == RHS will be <b>true</b>.
 * <p>If the LHS is a SlipPointer object (PTR or STRING) then the object
 *    pointed to must be deleted (for PTR if it is a user defined object and
 *    for STRING if it is a string). This decision is deferred to the
 *    SlipPointer <em>remove()</em> method. If the object is treated as a
 *    constant in SLIP, then <em>remove</em> does nothing. If the object is
 *    not a constant, the <em>remove()</em> deletes it. <em>remove()</em> is a
 *    proxy for <em>delete (void*)</em> within the context of SlipPointer.<p>
 * @param[in] X (bool) the right hand side of the assignment
 * @return reference to the current <em>SlipSequencer</em>
 */

/**
 * @fn  SlipIteration::operator=(const CHAR    X)
 * @brief Assign (<b>Y = X</b>) where Y is the current cell.
 * <p>For purposes of this section, <b>Y = X</b> refers to the left hand
 *    side (LHS) being replaced by the right hand side (RHS) of the
 *    assignment symbol (<b>=</b>). The LHS refers to the current object
 *    and the RHS refers to the input object.</p>
 * <p>Assignments are class invariant. Assignments do not preserve type.
 *    Under assignment the LHS <em>datum</em> field will be
 *    changed according to the RHS <em>datum</em> field and the LHS type
 *    field will be changed to correspond to the RHS.</p>
 * <p>When the assignment is complete, the RHS will not have changed and
 *    the LHS type will be CHAR and the value will be the same as the RHS.
 *    LHS == RHS will be <b>true</b>.
 * <p>If the LHS is a SlipPointer object (PTR or STRING) then the object
 *    pointed to must be deleted (for PTR if it is a user defined object and
 *    for STRING if it is a string). This decision is deferred to the
 *    SlipPointer <em>remove()</em> method. If the object is treated as a
 *    constant in SLIP, then <em>remove</em> does nothing. If the object is
 *    not a constant, the <em>remove()</em> deletes it. <em>remove()</em> is a
 *    proxy for <em>delete (void*)</em> within the context of SlipPointer.<p>
 * @param[in] X (CHAR) the right hand side of the assignment
 * @return reference to the current <em>SlipSequencer</em>
 */

/**
 * @fn SlipIteration::operator=(const DOUBLE  X)
 * @brief Assign (<b>Y = X</b>) where Y is the current cell.
 * <p>For purposes of this section, <b>Y = X</b> refers to the left hand
 *    side (LHS) being replaced by the right hand side (RHS) of the
 *    assignment symbol (<b>=</b>). The LHS refers to the current object
 *    and the RHS refers to the input object.</p>
 * <p>Assignments are class invariant. Assignments do not preserve type.
 *    Under assignment the LHS <em>datum</em> field will be
 *    changed according to the RHS <em>datum</em> field and the LHS type
 *    field will be changed to correspond to the RHS.</p>
 * <p>When the assignment is complete, the RHS will not have changed and
 *    the LHS type will be DOUBLE and the value will be the same as the RHS.
 *    LHS == RHS will be <b>true</b>.
 * <p>If the LHS is a SlipPointer object (PTR or STRING) then the object
 *    pointed to must be deleted (for PTR if it is a user defined object and
 *    for STRING if it is a string). This decision is deferred to the
 *    SlipPointer <em>remove()</em> method. If the object is treated as a
 *    constant in SLIP, then <em>remove</em> does nothing. If the object is
 *    not a constant, the <em>remove()</em> deletes it. <em>remove()</em> is a
 *    proxy for <em>delete (void*)</em> within the context of SlipPointer.<p>
 * @param[in] X (DOUBLE) the right hand side of the assignment
 * @return reference to the current <em>SlipSequencer</em>
 */

/**
 * @fn SlipIteration::operator=(const LONG    X)
 * @brief Assign (<b>Y = X</b>) where Y is the current cell.
 * <p>For purposes of this section, <b>Y = X</b> refers to the left hand
 *    side (LHS) being replaced by the right hand side (RHS) of the
 *    assignment symbol (<b>=</b>). The LHS refers to the current object
 *    and the RHS refers to the input object.</p>
 * <p>Assignments are class invariant. Assignments do not preserve type.
 *    Under assignment the LHS <em>datum</em> field will be
 *    changed according to the RHS <em>datum</em> field and the LHS type
 *    field will be changed to correspond to the RHS.</p>
 * <p>When the assignment is complete, the RHS will not have changed and
 *    the LHS type will be LONG and the value will be the same as the RHS.
 *    LHS == RHS will be <b>true</b>.
 * <p>If the LHS is a SlipPointer object (PTR or STRING) then the object
 *    pointed to must be deleted (for PTR if it is a user defined object and
 *    for STRING if it is a string). This decision is deferred to the
 *    SlipPointer <em>remove()</em> method. If the object is treated as a
 *    constant in SLIP, then <em>remove</em> does nothing. If the object is
 *    not a constant, the <em>remove()</em> deletes it. <em>remove()</em> is a
 *    proxy for <em>delete (void*)</em> within the context of SlipPointer.<p>
 * @param[in] X (LONG) the right hand side of the assignment
 * @return reference to the current <em>SlipSequencer</em>
 */

/**
 * @fn SlipIteration::operator=(const UCHAR   X)
 * @brief Assign (<b>Y = X</b>) where Y is the current cell.
 * <p>For purposes of this section, <b>Y = X</b> refers to the left hand
 *    side (LHS) being replaced by the right hand side (RHS) of the
 *    assignment symbol (<b>=</b>). The LHS refers to the current object
 *    and the RHS refers to the input object.</p>
 * <p>Assignments are class invariant. Assignments do not preserve type.
 *    Under assignment the LHS <em>datum</em> field will be
 *    changed according to the RHS <em>datum</em> field and the LHS type
 *    field will be changed to correspond to the RHS.</p>
 * <p>When the assignment is complete, the RHS will not have changed and
 *    the LHS type will be UCHAR and the value will be the same as the RHS.
 *    LHS == RHS will be <b>true</b>.
 * <p>If the LHS is a SlipPointer object (PTR or STRING) then the object
 *    pointed to must be deleted (for PTR if it is a user defined object and
 *    for STRING if it is a string). This decision is deferred to the
 *    SlipPointer <em>remove()</em> method. If the object is treated as a
 *    constant in SLIP, then <em>remove</em> does nothing. If the object is
 *    not a constant, the <em>remove()</em> deletes it. <em>remove()</em> is a
 *    proxy for <em>delete (void*)</em> within the context of SlipPointer.<p>
 * @param[in] X (UCHAR) the right hand side of the assignment
 * @return reference to the current <em>SlipSequencer</em>
 */

/**
 * @fn SlipIteration::operator=(const ULONG   X)
 * @brief Assign (<b>Y = X</b>) where Y is the current cell.
 * <p>For purposes of this section, <b>Y = X</b> refers to the left hand
 *    side (LHS) being replaced by the right hand side (RHS) of the
 *    assignment symbol (<b>=</b>). The LHS refers to the current object
 *    and the RHS refers to the input object.</p>
 * <p>Assignments are class invariant. Assignments do not preserve type.
 *    Under assignment the LHS <em>datum</em> field will be
 *    changed according to the RHS <em>datum</em> field and the LHS type
 *    field will be changed to correspond to the RHS.</p>
 * <p>When the assignment is complete, the RHS will not have changed and
 *    the LHS type will be ULONG and the value will be the same as the RHS.
 *    LHS == RHS will be <b>true</b>.
 * <p>If the LHS is a SlipPointer object (PTR or STRING) then the object
 *    pointed to must be deleted (for PTR if it is a user defined object and
 *    for STRING if it is a string). This decision is deferred to the
 *    SlipPointer <em>remove()</em> method. If the object is treated as a
 *    constant in SLIP, then <em>remove</em> does nothing. If the object is
 *    not a constant, the <em>remove()</em> deletes it. <em>remove()</em> is a
 *    proxy for <em>delete (void*)</em> within the context of SlipPointer.<p>
 * @param[in] X (ULONG) the right hand side of the assignment
 * @return reference to the current <em>SlipSequencer</em>
 */

/**
 * @fn SlipIteration::operator=(const PTR     X)
 * @brief Assign (<b>Y = X</b>) where Y is the current cell.
 * <p>For purposes of this section, <b>Y = X</b> refers to the left hand
 *    side (LHS) being replaced by the right hand side (RHS) of the
 *    assignment symbol (<b>=</b>). The LHS refers to the current object
 *    and the RHS refers to the input object.</p>
 * <p>Assignments are class invariant. Assignments do not preserve type.
 *    Under assignment the LHS <em>datum</em> field will be
 *    changed according to the RHS <em>datum</em> field and the LHS type
 *    field will be changed to correspond to the RHS.</p>
 * <p>When the assignment is complete, the RHS will not have changed and
 *    the LHS type will be PTR and the value will be the same as the RHS.
 *    LHS == RHS will be <b>true</b>.
 * <p>If the LHS is a SlipPointer object (PTR or STRING) then the object
 *    pointed to must be deleted (for PTR if it is a user defined object and
 *    for STRING if it is a string). This decision is deferred to the
 *    SlipPointer <em>remove()</em> method. If the object is treated as a
 *    constant in SLIP, then <em>remove</em> does nothing. If the object is
 *    not a constant, the <em>remove()</em> deletes it. <em>remove()</em> is a
 *    proxy for <em>delete (void*)</em> within the context of SlipPointer.<p>
 * @param[in] X (PTR) the right hand side of the assignment
 * @return reference to the current <em>SlipSequencer</em>
 */

/**
 * @fn SlipIteration::operator=(const string& X)
 * @brief Assign (<b>Y = X</b>) where Y is the current cell.
 * <p>For purposes of this section, <b>Y = X</b> refers to the left hand
 *    side (LHS) being replaced by the right hand side (RHS) of the
 *    assignment symbol (<b>=</b>). The LHS refers to the current object
 *    and the RHS refers to the input object.</p>
 * <p>Assignments are class invariant. Assignments do not preserve type.
 *    Under assignment the LHS <em>datum</em> field will be
 *    changed according to the RHS <em>datum</em> field and the LHS type
 *    field will be changed to correspond to the RHS.</p>
 * <p>When the assignment is complete, the RHS will not have changed and
 *    the LHS type will be STRING and the value will be the same as the RHS.
 *    LHS == RHS will be <b>true</b>.
 * <p>If the LHS is a SlipPointer object (PTR or STRING) then the object
 *    pointed to must be deleted (for PTR if it is a user defined object and
 *    for STRING if it is a string). This decision is deferred to the
 *    SlipPointer <em>remove()</em> method. If the object is treated as a
 *    constant in SLIP, then <em>remove</em> does nothing. If the object is
 *    not a constant, the <em>remove()</em> deletes it. <em>remove()</em> is a
 *    proxy for <em>delete (void*)</em> within the context of SlipPointer.<p>
 * @param[in] X (string&) the right hand side of the assignment
 * @return reference to the current <em>SlipSequencer</em>
 */

/**
 * @fn SlipIteration::operator=(const string* X)
 * @brief Assign (<b>Y = X</b>) where Y is the current cell.
 * <p>For purposes of this section, <b>Y = X</b> refers to the left hand
 *    side (LHS) being replaced by the right hand side (RHS) of the
 *    assignment symbol (<b>=</b>). The LHS refers to the current object
 *    and the RHS refers to the input object.</p>
 * <p>Assignments are class invariant. Assignments do not preserve type.
 *    Under assignment the LHS <em>datum</em> field will be
 *    changed according to the RHS <em>datum</em> field and the LHS type
 *    field will be changed to correspond to the RHS.</p>
 * <p>When the assignment is complete, the RHS will not have changed and
 *    the LHS type will be STRING and the value will be the same as the RHS.
 *    LHS == RHS will be <b>true</b>.
 * <p>If the LHS is a SlipPointer object (PTR or STRING) then the object
 *    pointed to must be deleted (for PTR if it is a user defined object and
 *    for STRING if it is a string). This decision is deferred to the
 *    SlipPointer <em>remove()</em> method. If the object is treated as a
 *    constant in SLIP, then <em>remove</em> does nothing. If the object is
 *    not a constant, the <em>remove()</em> deletes it. <em>remove()</em> is a
 *    proxy for <em>delete (void*)</em> within the context of SlipPointer.<p>
 * @param[in] X (string&) the right hand side of the assignment
 * @return reference to the current <em>SlipSequencer</em>
 */

/*****************************************
 *             Cell Equality             *
            ******************************************/
/**
 * @fn SlipIteration::operator==(const SlipReader& X)
 * @brief Compare the current cell to a SlipReader..
 * @param[in] X (SlipReaderr&) comparand
 * @return <b>true</b> if both the current list and current cell are equal
 */

/**
 * @fn SlipIteration::operator==(const SlipSequencer& X)
 * @brief Compare two <em>SlipSequencer</em>s.
 * @param[in] X (SlipSequencer&) comparand
 * @return <b>true</b> if both the current list and current cell are equal
 */

/**
 * @fn SlipIteration::operator==(const SlipCell& X)
 * @brief Compares the current cell with a <em>SlipCell</em>.
 * <p>If the type or values or the current cell and the input <em>SlipCell</em>
 *    are different then the result if <b>false</b>.</p>
 * @param[in] X (SlipCell&) comparand
 * @return <b>true</b> if the cells are equal
 */

/**
 * @fn SlipIteration::operator==(const bool    X)
 * @brief Compares the current cell with a bool.
 * <p>If the type or value of the current cell is not the same as the input
 *    then the result of the comparison is <b>false</b> otherwise <b>true</b>.
 *    </p>
 * @param[in] X (bool) X comparand
 * @return <b>true</b> if the type and values are the same
 */

/**
 * @fn SlipIteration::operator==(const CHAR    X)
 * @brief Compares the current cell with a CHAR.
 * <p>If the type or value of the current cell is not the same as the input
 *    then the result of the comparison is <b>false</b> otherwise <b>true</b>.
 *    </p>
 * @param[in] X (CHAR) X comparand
 * @return <b>true</b> if the type and values are the same
 */

/**
 * @fn SlipIteration::operator==(const DOUBLE  X)
 * @brief Compares the current cell with a DOUBLE.
 * <p>If the type or value of the current cell is not the same as the input
 *    then the result of the comparison is <b>false</b> otherwise <b>true</b>.
 *    </p>
 * @param[in] X (DOUBLE) X comparand
 * @return <b>true</b> if the type and values are the same
 */

/**
 * @fn SlipIteration::operator==(const LONG    X)
 * @brief Compares the current cell with a LONG.
 * <p>If the type or value of the current cell is not the same as the input
 *    then the result of the comparison is <b>false</b> otherwise <b>true</b>.
 *    </p>
 * @param[in] X (LONG) X comparand
 * @return <b>true</b> if the type and values are the same
 */

/**
 * @fn SlipIteration::operator==(const UCHAR   X)
 * @brief Compares the current cell with a UCHAR.
 * <p>If the type or value of the current cell is not the same as the input
 *    then the result of the comparison is <b>false</b> otherwise <b>true</b>.
 *    </p>
 * @param[in] X (UCHAR) X comparand
 * @return <b>true</b> if the type and values are the same
 */

/**
 * @fn SlipIteration::operator==(const ULONG   X)
 * @brief Compares the current cell with a ULONG.
 * <p>If the type or value of the current cell is not the same as the input
 *    then the result of the comparison is <b>false</b> otherwise <b>true</b>.
 *    </p>
 * @param[in] X (ULONG) X comparand
 * @return <b>true</b> if the type and values are the same
 */

/**
 * @fn SlipIteration::operator==(const PTR X)
 * @brief Compares the current cell with a user defined pointer object.
 * <p>If the type or value of the current cell is not the same as the input
 *    then the result of the comparison is <b>false</b> otherwise <b>true</b>.
 *    </p>
 * @param[in] X (PTR) X comparand
 * @return <b>true</b> if the type and values are the same
 */

/**
 * @fn SlipIteration::operator==(const string& X)
 * @brief Compares the current cell with a string.
 * <p>If the type or value of the current cell is not the same as the input
 *    then the result of the comparison is <b>false</b> otherwise <b>true</b>.
 *    </p>
 * @param[in] X (string&) X comparand
 * @return <b>true</b> if the type and values are the same
 */

/**
 * @fn SlipIteration::operator==(const string* X)
 * @brief Compares the current cell with a string.
 * <p>If the type or value of the current cell is not the same as the input
 *    then the result of the comparison is <b>false</b> otherwise <b>true</b>.
 *    </p>
 * @param[in] X (string*) X comparand
 * @return <b>true</b> if the type and values are the same
 */

/**
 * @fn SlipIteration::operator!=(const SlipSequencer&  X)
 * @brief Compare two <em>SlipSequencer</em>s.
 * @param[in] X (SlipSequencer&) comparand
 * @return <b>true</b> if either the current list or current cell are unequal
 */

/**
 * @fn SlipIteration::operator!=(const SlipCell& X)
 * @brief Compares the current cell with a <em>SlipCell</em>.
 * <p>If the type or values or the current cell and the input <em>SlipCell</em>
 *    are different then the result is <b>true</b>.</p>
 * @param[in] X (SlipCell&) comparand
 * @return <b>true</b> if the cells are unequal
 */

/**
 * @fn SlipIteration::operator!=(const bool    X)
 * @brief Compares the current cell with a bool.
 * <p>If the type or value of the current cell is not different as the input
 *    then the result of the comparison is <b>false</b> otherwise <b>true</b>.
 *    </p>
 * @param[in] X (bool) X comparand
 * @return <b>true</b> if the type and values are different
 */

/**
 * @fn SlipIteration::operator!=(const CHAR    X)
 * @brief Compares the current cell with a CHAR.
 * <p>If the type or value of the current cell is not different as the input
 *    then the result of the comparison is <b>false</b> otherwise <b>true</b>.
 *    </p>
 * @param[in] X (CHAR) X comparand
 * @return <b>true</b> if the type and values are different
 */

/**
 * @fn SlipIteration::operator!=(const DOUBLE  X)
 * @brief Compares the current cell with a DOUBLE.
 * <p>If the type or value of the current cell is not different as the input
 *    then the result of the comparison is <b>false</b> otherwise <b>true</b>.
 *    </p>
 * @param[in] X (DOUBLE) X comparand
 * @return <b>true</b> if the type and values are different
 */

/**
 * @fn SlipIteration::operator!=(const LONG    X)
 * @brief Compares the current cell with a LONG.
 * <p>If the type or value of the current cell is not different as the input
 *    then the result of the comparison is <b>false</b> otherwise <b>true</b>.
 *    </p>
 * @param[in] X (LONG) X comparand
 * @return <b>true</b> if the type and values are different
 */

/**
 * @fn SlipIteration::operator!=(const UCHAR   X)
 * @brief Compares the current cell with a UCHAR.
 * <p>If the type or value of the current cell is not different as the input
 *    then the result of the comparison is <b>false</b> otherwise <b>true</b>.
 *    </p>
 * @param[in] X (UCHAR) X comparand
 * @return <b>true</b> if the type and values are different
 */

/**
 * @fn SlipIteration::operator!=(const ULONG   X
 * @brief Compares the current cell with a ULONG.
 * <p>If the type or value of the current cell is not different as the input
 *    then the result of the comparison is <b>false</b> otherwise <b>true</b>.
 *    </p>
 * @param[in] X (ULONG) X comparand
 * @return <b>true</b> if the type and values are different
 */

/**
 * @fn SlipIteration::operator!=(const PTR X)
 * @brief Compares the current cell with a user defined pointer object.
 * <p>If the type or value of the current cell is not different as the input
 *    then the result of the comparison is <b>false</b> otherwise <b>true</b>.
 *    </p>
 * @param[in] X (PTR) X comparand
 * @return <b>true</b> if the type and values are different
 */

/**
 * @fn SlipIteration::operator!=(const string& X)
 * @brief Compares the current cell with a string.
 * <p>If the type or value of the current cell is not different as the input
 *    then the result of the comparison is <b>false</b> otherwise <b>true</b>.
 *    </p>
 * @param[in] X (string&) X comparand
 * @return <b>true</b> if the type and values are different
 */

/**
 * @fn SlipIteration::operator!=(const string* X)
 * @brief Compares the current cell with a string.
 * <p>If the type or value of the current cell is not different as the input
 *    then the result of the comparison is <b>false</b> otherwise <b>true</b>.
 *    </p>
 * @param[in] X (string*) X comparand
 * @return <b>true</b> if the type and values are different
 */

/*****************************************
 *           Manipulating Cells          *
                                   ******************************************/

/**
 * @fn SlipIteration::insLeft(SlipReader&       X)
 * @brief Insert the reader current cell to the left of the current cell.
 * @param[in] X (SlipReader&) value to be inserted
 * @return (SlipCell&) reference to the inserted cell
 */

/**
 * @fn SlipIteration::insLeft(SlipSequencer&    X)
 * @brief Insert a sequencer current cell to the left of the current cell.
 * @param[in] X (SlipReader&) value to be inserted
 * @return (SlipCell&) reference to inserted cell
 */

/**
 * @fn SlipIteration::insLeft(SlipCell&  X)
 * @brief Insert an input SlipCell object to the left of the current cell.
 * <p>The input object must be a valid source, and the current (<b>this</b>)
 *    object a valid destination for the insertion.</p>
 * @param[in] X (SlipCell&) SlipCell object reference
 * @return (SlipCell&) reference to inserted cell
 */

/**
 * @fn SlipIteration::insLeft(bool       X)
 * @brief Insert a bool to the left of the current cell.
 * <p>A <em>SlipDatum</em> cell with type bool containing the input value
 *    is inserted to the left of the current cell.</p>
 * @param[in] X (bool) value to be inserted
 * @return (SlipCell&) reference to inserted cell
 */

/**
 * @fn SlipIteration::insLeft(DOUBLE     X)
 * @brief Insert a DOUBLE to the left of the current cell.
 * <p>A <em>SlipDatum</em> cell with type DOUBLE containing the input value
 *    is inserted to the left of the current cell.</p>
 * @param[in] X (DOUBLE) value to be inserted
 * @return (SlipSequencer&) reference to the reader
 */

/**
 * @fn SlipIteration::insLeft(DOUBLE     X)
 * @brief Insert a DOUBLE to the left of the current cell.
 * <p>A <em>SlipDatum</em> cell with type DOUBLE containing the input value
 *    is inserted to the left of the current cell.</p>
 * @param[in] X (DOUBLE) value to be inserted
 * @return (SlipCell&) reference to inserted cell
 */

/**
 * @fn SlipIteration::insLeft(ULONG      X)
 * @brief Insert a ULONG to the left of the current cell.
 * <p>A <em>SlipDatum</em> cell with type ULONG containing the input value
 *    is inserted to the left of the current cell.</p>
 * @param[in] X (ULONG) value to be inserted
 * @return (SlipCell&) reference to inserted cell
 */

/**
 * @fn SlipIteration::insLeft(LONG       X)
 * @brief Insert a LONG to the left of the current cell.
 * <p>A <em>SlipDatum</em> cell with type LONG containing the input value
 *    is inserted to the left of the current cell.</p>
 * @param[in] X (LONG) value to be inserted
 * @return (SlipCell&) reference to inserted cell
 */

/**
 * @fn SlipIteration::insLeft(UCHAR      X)
 * @brief Insert a UCHAR to the left of the current cell.
 * <p>A <em>SlipDatum</em> cell with type UCHAR containing the input value
 *    is inserted to the left of the current cell.</p>
 * @param[in] X (UCHAR) value to be inserted
 * @return (SlipCell&) reference to inserted cell
 */

/**
 * @fn SlipIteration::insLeft(CHAR       X)
 * @brief Insert a CHAR to the left of the current cell.
 * <p>A <em>SlipDatum</em> cell with type CHAR containing the input value
 *    is inserted to the left of the current cell.</p>
 * @param[in] X (CHAR) value to be inserted
 * @return (SlipSequencer&) reference to the reader
 */

/**
 * @fn SlipIteration::insLeft(const PTR  X, const void* operation)
 * @brief Insert a user defined object to the left of the current cell.
 * <p>A <em>SlipDatum</em> cell with type PTR containing the input value
 *    is inserted to the left of the current cell.</p>
 * <p>The operations field provides a means to specify user defined operations.
 *    The pointer must be to a SlipOp defined class. The standard casting,
 *    unary, arithmetic, bitwise and logical operations must all be overridden.
 *    The result is that a user Defined object is treated as a atomic value
 *    in all computations.</p>
 * @param[in] X (PTR) X object to be inserted
 * @param[in] operation (void*) SlipOp derived User Defined operations
 * @return (SlipCell&) reference to inserted cell
 */

/**
 * @fn SlipIteration::insLeft(const string& X, bool constFlag)
 * @brief Insert a string to the left of the current cell.
 * <p>A <em>SlipDatum</em> cell with type string containing the input value
 *    is inserted to the left of the current cell.</p>
 * @param[in] X (string&) value to be inserted
 * @param[in] constFlag <b>true</b> string is a <b>const</b>, <b>false</b> otherwise
 * @return (SlipCell&) reference to inserted cell
 */

/**
 * @fn SlipIteration::insLeft(const string* X, bool constFlag)
 * @brief Insert a string to the left of the current cell.
 * <p>A <em>SlipDatum</em> cell with type string containing the input value
 *    is inserted to the left of the current cell.</p>
 * @param[in] X (string*) value to be inserted
 * @param[in] constFlag <b>true</b> string is a <b>const</b>, <b>false</b> otherwise
 * @return (SlipCell&) reference to inserted cell
 */
                                                // List Cells must be unlinked

/**
 * @fn SlipIteration::insRight(SlipReader&       X)
 * @brief Insert the reader current cell to the right of the current cell.
 * @param[in] X (SlipReader&) value to be inserted
 * @return (SlipCell&) reference to inserted cell
 */

/**
 * @fn SlipIteration::insRight(SlipSequencer&    X)
 * @brief Insert a sequencer current cell to the right of the current cell.
 * @param[in] X (SlipReader&) value to be inserted
 * @return (SlipCell&) reference to inserted cell
 */

/**
 * @fn SlipIteration::insRight(SlipCell&  X)
 * @brief Insert an input SlipCell object to the right of the current cell.
 * <p>The input object must be a valid source, and the current (<b>this</b>)
 *    object a valid destination for the insertion.</p>
 * @param[in] X (SlipCell&) SlipCell object reference
 * @return (SlipCell&) reference to inserted cell
 */

/**
 * @fn SlipIteration::insRight(bool       X)
 * @brief Insert a bool to the right of the current cell.
 * <p>A <em>SlipDatum</em> cell with type bool containing the input value
 *    is inserted to the right of the current cell.</p>
 * @param[in] X (bool) value to be inserted
 * @return (SlipCell&) reference to inserted cell
 */

/**
 * @fn SlipIteration::insRight(DOUBLE       X)
 * @brief Insert a DOUBLE to the right of the current cell.
 * <p>A <em>SlipDatum</em> cell with type DOUBLE containing the input value
 *    is inserted to the right of the current cell.</p>
 * @param[in] X (DOUBLE) value to be inserted
 * @return (SlipCell&) reference to inserted cell
 */

/**
 * @fn SlipIteration::insRight(ULONG      X)
 * @brief Insert a ULONG to the right of the current cell.
 * <p>A <em>SlipDatum</em> cell with type ULONG containing the input value
 *    is inserted to the right of the current cell.</p>
 * @param[in] X (ULONG) value to be inserted
 * @return (SlipCell&) reference to inserted cell
 */

/**
 * @fn SlipIteration::insRight(LONG       X)
 * @brief Insert a LONG to the right of the current cell.
 * <p>A <em>SlipDatum</em> cell with type LONG containing the input value
 *    is inserted to the right of the current cell.</p>
 * @param[in] X (LONG) value to be inserted
 * @return (SlipCell&) reference to inserted cell
 */

/**
 * @fn SlipIteration::insRight(UCHAR      X)
 * @brief Insert a UCHAR to the left of the current cell.
 * <p>A <em>SlipDatum</em> cell with type UCHAR containing the input value
 *    is inserted to the left of the current cell.</p>
 * @param[in] X (UCHAR) value to be inserted
 * @return (SlipCell&) reference to inserted cell
 */

/**
 * @fn SlipIteration::insRight(CHAR       X)
 * @brief Insert a CHAR to the left of the current cell.
 * <p>A <em>SlipDatum</em> cell with type CHAR containing the input value
 *    is inserted to the left of the current cell.</p>
 * @param[in] X (CHAR) value to be inserted
 * @return (SlipCell&) reference to inserted cell
 */

/**
 * @fn SlipIteration::insRight(const PTR X, const void* operation)
 * @brief Insert a user defined object to the left of the current cell.
 * <p>A <em>SlipDatum</em> cell with type PTR containing the input value
 *    is inserted to the left of the current cell.</p>
 * <p>The operations field provides a means to specify user defined operations.
 *    The pointer must be to a SlipOp defined class. The standard casting,
 *    unary, arithmetic, bitwise and logical operations must all be overridden.
 *    The result is that a user Defined object is treated as a atomic value
 *    in all computations.</p>
 * @param[in] X (PTR) X object to be inserted
 * @param[in] operation (void*) SlipOp derived User Defined operations
 * @return (SlipCell&) reference to inserted cell
 */

/**
 * @fn SlipIteration::insRight(const string& X, bool constFlag)
 * @brief Insert a string to the left of the current cell.
 * <p>A <em>SlipDatum</em> cell with type string containing the input value
 *    is inserted to the left of the current cell.</p>
 * @param[in] X (string&) value to be inserted
 * @param[in] constFlag <b>true</b> string is a <b>const</b>, <b>false</b> otherwise
 * @return (SlipCell&) reference to inserted cell
 */

/**
 * @fn SlipIteration::insRight(const string*    X, bool constFlag)
 * @brief Insert a string to the left of the current cell.
 * <p>A <em>SlipDatum</em> cell with type string containing the input value
 *    is inserted to the left of the current cell.</p>
 * @param[in] X (string*) value to be inserted
 * @param[in] constFlag <b>true</b> string is a <b>const</b>, <b>false</b> otherwise
 * @return (SlipCell&) reference to inserted cell
 */

/**
 * @fn SlipIteration::moveLeft(SlipCell& X)
 * @brief Move the input cell to the left of the current cell.
 * <p>Check to see if the source and destination of the move are valid. If
 *    valid, then:</p>
 * <ul>
 *    <li>If the source is a list, then the entire list is moved to the
 *        left of the current object and the list is empty after the move.</li>
 *    <li>If the source is linked, then it is unlinked and moved to the
 *        left of the current object.</li>
 *    <li>If the source is unlinked, then it is inserted to the left of
 *        the current object.</li>
 * </ul>
 * <p>The sequencer current cell reference is unchanged.</p>
 * <p><b>ERRORS</b></p>
 * <ul>
 *    <li><b>E2006</b> Cell can not be a local temporary. Create with 'new'.</li>
 *    <li><b>E2013</b> Cell must be in a list. Cell is not in a list.</li>
 *    <li><b>E3019</b> Cell has been deleted. Reference to cell is stale.</li>
 *    <li><b>E3022</b> Can not insert a non-AVSL cell into a list.</li>
 * </ul>
 * @param[in] X (SlipCell&) to be inserted to the left
 * @return (SlipCell&) reference to the current cell
 */

/**
 * @brief Move the reader current cell to the left of the current cell.
 * <ul>
 *    <li>If the source is a list, then the entire list is moved to the
 *        left of the current object and the list is empty after the move.
 *        The input SlipSequencer references an empty list.</li>
 *    <li>If the source is linked, then it is unlinked and moved to the
 *        left of the current object. The reader is updated to point to the
 *        cell preceding the moved cell.</li>
 * </ul>
 * <p>The sequencer current cell reference is unchanged.</p>
 * @param[in] X (SlipReader&) current cell to be inserted to the left
 * @return (SlipCell&) reference to the current cell
 */

/**
 * @brief Move the sequencer current cell to the left of the current cell.
 * <ul>
 *    <li>If the source is a list, then the entire list is moved to the
 *        left of the current object and the list is empty after the move.
 *        The input SlipSequencer references an empty list.</li>
 *    <li>If the source is linked, then it is unlinked and moved to the
 *        left of the current object. The input sequencer current cell is
 *        changed to the preceding cell.</li>
 * </ul>
 * @param[in] X (SlipSequencer&) current cell to be inserted to the left
 * @return (SlipCell&) reference to the current cell
 */

/**
 * @fn SlipIteration::moveRight(SlipCell& X)
 * @brief Move the input cell to the right of the current cell.
 * <p>Check to see if the source and destination of the move are valid. If
 *    valid, then:</p>
 * <ul>
 *    <li>If the source is a list, then the entire list is moved to the
 *        right of the current object and the list. The list is empty after
 *        the move.</li>
 *    <li>If the source is linked, then it is unlinked and moved to the
 *        right of the current object.</li>
 *    <li>If the source is unlinked, then it is inserted to the right of
 *        the current object.</li>
 * </ul>
 * <p><b>ERRORS</b></p>
 * <ul>
 *    <li><b>E2006</b> Cell can not be a local temporary. Create with 'new'.</li>
 *    <li><b>E2013</b> Cell must be in a list. Cell is not in a list.</li>
 *    <li><b>E3019</b> Cell has been deleted. Reference to cell is stale.</li>
 *    <li><b>E3022</b> Can not insert a non-AVSL cell into a list.</li>
 * </ul>
 * @param[in] X (SlipCell&) to be inserted to the right
 * @return (SlipCell&) reference to the current cell
 */

/**
 * @brief Move the reader current cell to the right of the current cell.
 * <ul>
 *    <li>If the source is a list, then the entire list is moved to the
 *        right of the current object and the list is empty after the move.
 *        The input SlipReader references an empty list.</li>
 *    <li>If the source is linked, then it is unlinked and moved to the
 *        right of the current object. The reader is updated to point to the
 *        cell preceding the moved cell.</li>
 * </ul>
 * <p>The sequencer current cell reference is unchanged.</p>
 * @param[in] X (SlipReader&) cell to be inserted to the right
 * @return (SlipCell&) reference to the current cell
 */

/**
 * @brief Move the sequencer current cell to the right of the current cell.
 * <ul>
 *    <li>If the source is a list, then the entire list is moved to the
 *        right of the current object and the list is empty after the move.
 *        The input SlipSequencer references an empty list.</li>
 *    <li>If the source is linked, then it is unlinked and moved to the
 *        left of the current object. The input sequencer current cell is
 *        changed to the preceding cell.</li>
 * </ul>
 * @param[in] X (SlipSequencer&) current cell to be inserted to the left
 * @return (SlipCell&) reference to the sequencer
 */


/**
 * @fn SlipIteration::moveListLeft(SlipCell& X)
 * @brief Move the input list to the left of the current cell.
 * <p>Check to see if the source and destination of the move are valid. If
 *    valid, then:</p>
 * <ul>
 *    <li>The entire list is moved to the right of the current cell and the
 *        list is empty after the move.</li>
 *    <li>If the input is a list header, then the header has an empty list.</li>
 *    <li>If the list is a sublist then the sublist reference is to an
 *        empty list.</li>
 * </ul>
 * <p>The sequencer current cell reference is unchanged.</p>
 * @sa slip::SlipCell::moveListLeft()
 * @param[in] X (SlipCell&) list to be inserted to the left
 * @return (SlipCell&) reference to the current cell
 */

/**
 * @brief Move the reader current list to the left of the current cell.
 * <ul>
 *    <li>The entire list is moved to the left of the current object and
 *        the list is empty after the move. The input SlipReader current
 *        cell and current list are the same.</li>
 * </ul>
 * <p>The sequencer current cell reference is unchanged.</p>
 * @sa slip::SlipCell::moveListLeft()
 * @param[in] X (SlipReader&) current list to be inserted to the left
 * @return (SlipCell&) reference to the current cell
 */

/**
 * @brief Move the sequencer current list to the left of the current cell.
 * <ul>
 *    <li>The entire list is moved to the left of the current object and
 *        the list is empty after the move. The input SlipSequencer
 *        references an empty list.</li>
 *    <li>If the input is not a list then post an error message. The
 *        input sequencer current cell is unchanged.</li>
 * </ul>
 * <p><b>ERRORS</b></p>
 * <ul>
 *    <li><b>E1034</b> Invalid Operation for Data Cell Type.</li>
 * </ul>
 * @sa slip::SlipCell::moveListLeft()
 * @param[in] X (SlipSequencer&) current list to be inserted to the left
 * @return (SlipCell&) reference to the current cell
 */

/**
 * @fn SlipIteration::moveListRight(SlipCell& X)
 * @brief Move the input cell to the right of the current cell.
 * <p>Check to see if the source and destination of the move are valid. If
 *    valid, then:</p>
 * <ul>
 *    <li>The entire list is moved to the right of the current cell and the
 *        list is empty after the move.</li>
 *    <li>If the input is a list header, then the header has an empty list.</li>
 *    <li>If the list is a sublist then the sublist reference is to an
 *        empty list.</li>
 * </ul>
 * @sa slip::SlipCell::moveListRight()
                      *
 * @param[in] X (SlipCell&) to be inserted to the right
 * @return (SlipCell&) reference to the current cell
 */

/**
 * @brief Move the reader current cell to the right of the current cell.
 * <ul>
 *    <li>The entire list is moved to the right of the current object and
 *        the list is empty after the move. The input SlipReader references
 *        an empty list and the current cell is set to the list header.</li>
 * </ul>
 * <p>The sequencer current cell reference is unchanged.</p>
 * @sa slip::SlipCell::moveListRight()
 * @param[in] X (SlipReader&) current cell to be inserted to the left
 * @return (SlipCell&) reference to the current cell
 */

/**
 * @brief Move the sequencer current cell to the right of the current cell.
 * <ul>
 *    <li>The entire list is moved to the right of the current object and
 *        the list is empty after the move. The input SlipSequencer
 *        references an empty list.</li>
 *    <li>If the input is not a list then post an error message. The
 *        input sequencer current cell is unchanged.</li>
 * </ul>
 * <p><b>ERRORS</b></p>
 * <ul>
 *    <li><b>E1034</b> Invalid Operation for Data Cell Type.</li>
 * </ul>
 * @sa slip::SlipCell::moveListLeft()
 * @param[in] X (SlipSequencer&) current cell to be inserted to the left
 * @return (SlipCell&) reference to the sequencer
 */


/**
 * @fn SlipIteration::replace(const SlipReader& X)
 * @brief Replace the current cell with the current cell of the input reader.
 * <p>If the current list is empty or the current cell is the current list,
 *    issue a diagnostic message.</p>
 * <p><b>ERRORS</b></p>
 * <ul>
 *    <li><b>E2015</b>Header not allowed in replace method</li>
 * </ul>
 * @param[in] X (SlipSequencer&) replacement value
 * @return (SlipCell&) reference to changed cell
 */

/**
 * @fn SlipIteration::replace(const SlipSequencer& X)
 * @brief Replace the current cell with the current cell of the input sequencer.
 * <p>If the current list is empty or the current cell is the current list,
 *    issue a diagnostic message.</p>
 * <p><b>ERRORS</b></p>
 * <ul>
 *    <li><b>E2015</b>Header not allowed in replace method</li>
 * </ul>
 * @param[in] X (SlipSequencer&) replacement value
 * @return (SlipCell&) reference to changed cell
 */

/**
 * @fn SlipIteration::replace(const SlipCell& X)
 * @brief Replace the current cell with the input value.
 * <p>If the current list is empty or the current cell is the current list,
 *    issue a diagnostic message.</p>
 * <p><b>ERRORS</b></p>
 * <ul>
 *    <li><b>E2015</b>Header not allowed in replace method</li>
 * </ul>
 * @param[in] X (SlipCell&) replacement value
 * @return (SlipCell&) reference to changed cell
 */

/**
 * @fn SlipIteration::replace(bool X)
 * @brief Replace the current cell with the input value.
 * <p>If the current list is empty or the current cell is the current list,
 *    issue a diagnostic message.</p>
 * <p><b>ERRORS</b></p>
 * <ul>
 *    <li><b>E2015</b>Header not allowed in replace method</li>
 * </ul>
 * @param[in] X (bool) replacement value
 * @return (SlipCell&) reference to changed cell
 */

/**
 * @fn SlipIteration::replace(CHAR X)
 * @brief Replace the current cell with the input value.
 * <p>If the current list is empty or the current cell is the current list,
 *    issue a diagnostic message.</p>
 * <p><b>ERRORS</b></p>
 * <ul>
 *    <li><b>E2015</b>Header not allowed in replace method</li>
 * </ul>
 * @param[in] X (CHAR) replacement value
 * @return (SlipCell&) reference to changed cell
 */

/**
 * @fn SlipIteration::replace(UCHAR X)
 * @brief Replace the current cell with the input value.
 * <p>If the current list is empty or the current cell is the current list,
 *    issue a diagnostic message.</p>
 * <p><b>ERRORS</b></p>
 * <ul>
 *    <li><b>E2015</b>Header not allowed in replace method</li>
 * </ul>
 * @param[in] X (UCHAR) replacement value
 * @return (SlipCell&) reference to changed cell
 */

/**
 * @fn SlipIteration::replace(LONG X)
 * @brief Replace the current cell with the input value.
 * <p>If the current list is empty or the current cell is the current list,
 *    issue a diagnostic message.</p>
 * <p><b>ERRORS</b></p>
 * <ul>
 *    <li><b>E2015</b>Header not allowed in replace method</li>
 * </ul>
 * @param[in] X (LONG) replacement value
 * @return (SlipCell&) reference to changed cell
 */

/**
 * @fn SlipIteration::replace(ULONG X)
 * @brief Replace the current cell with the input value.
 * <p>If the current list is empty or the current cell is the current list,
 *    issue a diagnostic message.</p>
 * <p><b>ERRORS</b></p>
 * <ul>
 *    <li><b>E2015</b>Header not allowed in replace method</li>
 * </ul>
 * @param[in] X (ULONG) replacement value
 * @return (SlipCell&) reference to changed cell
 */

/**
 * @fn SlipIteration::replace(DOUBLE X)
 * @brief Replace the current cell with the input value.
 * <p>If the current list is empty or the current cell is the current list,
 *    issue a diagnostic message.</p>
 * <p><b>ERRORS</b></p>
 * <ul>
 *    <li><b>E2015</b>Header not allowed in replace method</li>
 * </ul>
 * @param[in] X (DOUBLE) replacement value
 * @return (SlipCell&) reference to changed cell
 */

/**
 * @fn SlipIteration::replace(const PTR X, const void* operation) {
 * @brief Replace the current cell with the input value.
 * <p>If the current list is empty or the current cell is the current list,
 *    issue a diagnostic message.</p>
 * <p>The operations field provides a means to specify user defined operations.
 *    The pointer must be to a SlipOp defined class. The standard casting,
 *    unary, arithmetic, bitwise and logical operations must all be overridden.
 *    The result is that a user Defined object is treated as a atomic value
 *    in all computations.</p>
 * <p><b>ERRORS</b></p>
 * <ul>
 *    <li><b>E2015</b>Header not allowed in replace method</li>
 * </ul>
 * @param[in] X (PTR) X object to be inserted
 * @param[in] operation (void*) SlipOp derived User Defined operations
 * @return (SlipCell&) reference to changed cell
 */

/**
 * @fn SlipIteration::replace(const string& X, bool constFlag)
 * @brief Replace the current cell with the input value.
 * <p>The sequencer will point to the new cell.</p>
 * @param[in] X (string&) replacement value
 * @param[in] constFlag <b>true</b> string and object referenced not copied
 * @return (SlipCell&) reference to changed cell
 */

/**
 * @fn SlipIteration::replace(const string* X, bool constFlag)
 * @brief Replace the current cell with the input value.
 * <p>The sequencer will point to the new cell.</p>
 * @param[in] X (string*) replacement value
 * @param[in] constFlag <b>true</b> string and object referenced not copied
 * @return (SlipCell&) reference to changed cell
 */

/**
 * @brief Unlink the current cell from a list and fixup sequencer.
 * <p>The current cell is unlinked and the sequencer is advanced one cell
 *    to the left.</p>
 * @return (SlipCell&) a reference to the unlinked cell.
 */

/*****************************************
 *             Miscellaneous              *
                                     ******************************************/


/**
 * @fn SlipIteration::getClassType() const
 * @brief Return current cell class type.
 * <p>Valid cell types are:</p>
 * | Cell  Type |   data type   |
 * | :--------- | :------------ |
 * | <b>eUNDEFINED</b>   | UNDEFINED     |
 * | <b>eBOOL</b>        | bool          |
 * | <b>eCHAR</b>        | CHAR          |
 * | <b>eUCHAR</b>       | UCHAR         |
 * | <b>eLONG</b>        | LONG          |
 * | <b>eULONG</b>       | ULONG         |
 * | <b>eDOUBLE</b>      | DOUBLE        |
 * | <b>eSUBLIST</b>     | SUBLIST       |
 * | <b>eSTRING</b>      | STRING        |
 * | <b>ePTR</b>         | PTR           |
 * | <b>eREADER</b>      | READER        |
 * | <b>eHEADER</b>      | HEADER        |
                                      *
 * @return class type of the object
 * @return (ClassType) class type
 */

/**
 * @fn SlipIteration::getName() const
 * @brief ASCII name of current cell
 * @return (string) ASCII name of current cell
 */

/**
 * @fn SlipIteration::isData() const
 * @brief <b>true</b> if current cell is a <em>SlipDatum</em> cell.
 * @return (bool) <b>true</b> if current cell is a <em>SlipDatum</em> cell.
 */

/**
 * @fn SlipIteration::isDiscrete() const
 * @brief <b>true</b> if current cell cell contains a discrete number.
 * @return (bool) <b>true</b> if current cell contains a discrete number.
 */

/**
 * @fn SlipIteration::isHeader() const
 * @brief <b>true</b> if current cell is a <em>SlipHeader</em> cell.
 * @return (bool) <b>true</b> if current cell is a <em>SlipHeader</em> cell
 */

/**
 * @fn SlipIteration::isName() const
 * @brief <b>true</b> if current cell is a <em>SlipSublist</em> cell.
 * @return (bool) <b>true</b> if current cell is a <em>SlipSublist</em> cell
 */

/**
 * @fn SlipIteration::isNumber() const
 * @brief <b>true</b> if current cell is a numeric type.
 * @return (bool) <b>true</b> if current cell is a numeric type
 */

/**
 * @fn SlipIteration::isPtr() const
 * @brief <b>true</b> if current cell is a user defined type.
 * @return (bool) <b>true</b> if current cell is a user defined type
 */

/**
 * @fn SlipIteration::isReal() const
 * @brief <b>true</b> if current cell contains a real number.
 * @return (bool) <b>true</b> if current cell contains a real number.
 */

/**
 * @fn SlipIteration::isString() const
 * @brief <b>true</b> if current cell contains a real number.
 * @return (bool) <b>true</b> if current cell contains a real number.
 */

/**
 * @fn SlipIteration::isSublist() const
 * @brief <b>true</b> if current cell is a <em>SlipSublist</em> cell.
 * @return (bool) <b>true</b> if current cell is a <em>SlipSublist</em> cell
 */

/*****************************************
 *    Miscellaneous Sequencer Methods    *
                                                ******************************************/
/**
 * @brief Change current cell to reference the input SlipCell.
 * <p>The input SlipCell is made the current reference of the
 *    SlipSequencer. If the input SlipCell is illegal, then
 *    the SlipSequencer link will point to an illegally
 *    constructed SlipDatum. Any operations with this cell
 *    should yield an error.</p>
 * <p><b>ERRORS</b></p>
 * <ul>
 *    <li><b>E2006</b> Cell can not be a local temporary. Create with 'new'.</li>
 *    <li><b>E2013</b> Cell must be in a list. Cell is not in a list.</li>
 *    <li><b>E3019</b> Slip Cell has been deleted and can not be used.</li>
 * </ul>
 * @param[in] X (SlipCell&) new current cell reference
 * @return (SlipSequencer&) reference to the sequencer
 */

/**
 * @fn SlipIteration::currentCell() const
 * @brief Reference the current cell.
 * @return (SlipCell&) current cell
 */

/*****************************************
 *         Miscellaneous Methods         *
                                                  ******************************************/

/**
 * @fn SlipIteration::getLeftLink()  const
 * @brief Return the leftLink.
 * @return Return the leftLink
 */

/**
 * @fn SlipIteration::getRightLink()
 * @brief Return the rightLink.
 * @return Return the rightLink
 */

/**
 * @fn SlipIteration::toString()
 * @brief The pretty-printed value of the current cell.
 * @return (string) pretty-printed current cell value
 */

/**
 * @fn SlipIteration::dump()
 * @brief Ugly representation of the current cell.
 * @return (string) uglified cell representation
 */


   class SlipSequencer : public SlipIterator {
    /*********************************************************
    *                    Private Functions                   *
    **********************************************************/
    private:
       SlipCell* link;                                //!< link to current cell

       SlipSequencer();

    /*********************************************************
    *                   Protected Functions                  *
    **********************************************************/
    protected:

   /*********************************************************
   *                    Public Functions                    *
   **********************************************************/
    public:
       SlipSequencer(SlipHeader& header);
       SlipSequencer(SlipSublist& sublist);
       SlipSequencer(const SlipHeader& header);
       SlipSequencer(const SlipSublist& sublist);
       SlipSequencer(const SlipSequencer& orig) { link = orig.link; };
       SlipSequencer(const SlipCell& orig) { };
       virtual ~SlipSequencer();
            /*****************************************
            *            Sequencer Iteration         *
            ******************************************/
      virtual SlipCell&     advanceLEL();                                          // advance linear element left,  true if element found
      virtual SlipCell&     advanceLER();                                          // advance linear element right, true if element found
      virtual SlipCell&     advanceLNL();                                          // advance linear name left,     true if element found
      virtual SlipCell&     advanceLNR();                                          // advance linear name right,    true if element found
      virtual SlipCell&     advanceLWL();                                          // advance linear word right,    true if element found
      virtual SlipCell&     advanceLWR();                                          // advance linear word right,    true if element found
      virtual SlipCell&     advanceSEL();                                          // advance structural element left,  true if element found
      virtual SlipCell&     advanceSER();                                          // advance structural element right, true if element found
      virtual SlipCell&     advanceSNL();                                          // advance structural name left,     true if element found
      virtual SlipCell&     advanceSNR();                                          // advance structural name right,    true if element found
      virtual SlipCell&     advanceSWL();                                          // advance structural word right,    true if element found
      virtual SlipCell&     advanceSWR();                                          // advance structural word right,    true if element found

            /*****************************************
            *         Sequencer Assignment          *
            ******************************************/
     
            /*****************************************
            *   Miscellaneous Sequencer Methods     *
            ******************************************/
       virtual SlipCell& currentCell() const { return *link;  }               // Return current SLIP cell being referenced
       virtual string    toString()    const { return link->toString(); }     // Pretty-print the cell
       virtual string    dump()        const { return link->SlipCellBase::dump(); }    // uglify the cell
 
            /*****************************************
             *             Miscellaneous             *
            ******************************************/
       virtual SlipCell&     deleteCell();                                    // delete the current cell
   }; // class SlipSequencer

}; // namespace slip

#endif	/* SLIPREADER_H */

