 /*Copyright (C) 2014 Arthur I. Schwarz

   This file is part of the C++ SLIP library.  This library is free
   software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the
   Free Software Foundation; either version 3, or (at your option)
   any later version.
  
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
  
   Under Section 7 of GPL version 3, you are granted additional
   permissions described in the GCC Runtime Library Exception, version
   3.1, as published by the Free Software Foundation.
  
   You should have received a copy of the GNU General Public License and
   a copy of the GCC Runtime Library Exception along with this program;
   see the files COPYING3 and COPYING.RUNTIME respectively.  If not, see
   <http:www.gnu.org/licenses/>.
 */

/*
  * Maintenance:
  * Compile: flex --noline slip.lex
  *
  * @note:
  *   1. As of 23 Jun 2023 the header file SlipFlex.h can not be
  *      be included. SlipFlex.h contains undef's and Flex puts
  *      the header file in a position to cause all the defines
  *      to be undefined.
  *   2. As of 23 Jun 2023 a separate file, Scanner.h, must be
  *      constructed and included to contain the scanner interface.
  *      The scanner interface routines can not be inclulded in
  *      SlipFlex.h (see 1. above).
  *   3. As of 2 Jul 2023 the option '%optin nounistd' does not work.
  *      '#include <unistd.h>' must be commented out in the generated 
  *      header (SlipFlex.h) files.
 */
 
/*****************************************************/
/*                   Flex Options                    */
/*****************************************************/
%option noline
%option never-interactive
%option nounistd
%option noyywrap
%option prefix="Slip"
%option header-file="SlipFlex.h"
%option outfile="SlipFlex.cpp"

/*****************************************************/
/*             Discarded Flex Options                */
/*****************************************************/
/*%option c++*/
/*%option bison-bridge*/
/*%option bison-locations*/


%{
# include <cerrno>
# include <cstddef>
# include <cstdlib>
# include <cstring>
# include <ctype.h>
# include <string>
# include <stdint.h>

/*****************************************************/
/*        Redefine values from namespace slip        */
/*****************************************************/
# undef BOOL
# undef CHAR
# undef FLOAT
# undef STRING
# undef UCHAR
# undef UINTEGER
# undef INTEGER
# undef SLIPCELL
# undef SLIPDATUM
# undef SLIPHEADER
# undef SLIPSUBLIST

//# include "SlipFlex.h"                        // Can not be included
# include "SlipParser.h"

using namespace std;
using namespace slip;

#define YY_DECL int Sliplex()
                           
#undef DEBUG
# ifdef DEBUG_IO
#   define DEBUG(flag, stmt) if (flag) stmt
# else
   #define DEBUG(flag, stmt)
# endif

# ifdef __cplusplus
# define isatty(arg) 0;
# endif

# define ATOF myatof
double myatof(char * const);

#define TOKEN yylval
#define MAX_STR_CONST 1025

   using namespace std;
   
   bool  lexerDebugFlag = false;

   string getFileName();                           /* get current file name */

   int   col    = yylloc.first_column;             /* start column of token */
   int   lineno = yylloc.first_line;               /* line number of token  */
   

%}
  
%x  COMMENT EOLCOMMENT STRING_CHAR

WHITESPACE  [[:space:]]{-}[\n]
EXP         [eE][+-]?[[:digit:]]+
CHAR        [cC]
LONG        [lL]
SIGN        [+|-]?
SIGNED      ({CHAR}|{LONG})
UNSIGNED    ([uU]({CHAR}|{LONG})?)
DECIMAL     {SIGN}[1-9][[:digit:]]*
HEXADECIMAL 0[xX][[:xdigit:]]+
OCTAL       0[0-7]* 
FLOAT1      {SIGN}[[:digit:]]+\.[[:digit:]]*{EXP}?
FLOAT2      {SIGN}\.[[:digit:]]+{EXP}?
FLOAT3      {SIGN}[[:digit:]]+{EXP}
FOLLOW      [,;:\(\)<>/{}|[:cntrl:]| ]
HEXFLOAT    0[xX]1.[:xdigit:]+[[p|P][+|-][0-9]+]?
NAMEEND     ([[:graph:]_\$]{-}[,;:\(\)<>/{}])
NAMESTART   [[:alpha:]_\$]
NAME        {NAMESTART}{NAMEEND}*
FALSE       [fF][aA][lL][sS][eE]
TRUE        [tT][rR][uU][eE]
%%
     bool stringFlag = false;                        /* false; quoted char    */
     string *string_buf = NULL;                      /* input string buffer   */

,                                { return yyCOMMA;  }
;                                { return yySEMI;   }
:                                { return yyCOLON;  }
\(                               { return yyLPARN;  }
\)                               { return yyRPARN;  }
\{                               { return yyLBRACE; }
\}                               { return yyRBRACE; }
\<                               { return yyLANGLE; }
\>                               { return yyRANGLE; }
#                                { DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d %c                  %s\n", lineno, col, yytext[0], yytext);) 
                                   col++; 
                                   return yyHASH; }
\\a                              { DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d \\a                %s\n", lineno, col, yytext);) col++; 
                                   yylval.CHAR = '\a';
                                   return yyBELL;}
\\b                              { DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d \\b                %s\n", lineno, col, yytext);) col++;
                                   yylval.CHAR = '\b';
                                   return yyBACK;}
\\f                              { DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d \\f                %s\n", lineno, col, yytext);) col++;
                                   yylval.CHAR = '\f';
                                   return yyFF;}
\\n                              { DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d \\n                %s\n", lineno, col, yytext);) col++;
                                   yylval.CHAR = '\n';
                                   return yyLF;}
\\r                              { DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d \\r                %s\n", lineno, col, yytext);) col++;
                                   yylval.CHAR = '\r';
                                   return yyCR;}
\\t                              { DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d \\t                %s\n", lineno, col, yytext);) col++;
                                   yylval.CHAR = '\t';
                                   return yyHT;}
\\.                              { DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d \\.                %s\n", lineno, col, yytext);) col++;
                                   yylval.CHAR = '.';
                                   return yyPERIOD;}
\"  |
\'                               { string whoWeAre = (yytext[0] == '"')? "STRING": "CHAR  ";
                                   DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d BEGIN %-s       %s\n", lineno, col, whoWeAre.c_str(), yytext);) 
                                   col++;
                                   string_buf = new string();
                                   stringFlag = (yytext[0] == '"')? true: false;
                                   BEGIN(STRING_CHAR); }
"/*"                             { DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d /* BEGIN COMMENT   %s\n", lineno, col, yytext);) col++; 
                                   BEGIN(COMMENT); }
""                               { DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d EOL COMMENT        %s\n", lineno, col, yytext);) col++; 
                                   BEGIN(EOLCOMMENT); }
     
{WHITESPACE}+                    { col += yyleng; }
\n                               { DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d \\n\n", lineno, col); lineno++; col=1;) }
{OCTAL}/{FOLLOW}                 { DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d OCTAL              %s\n", lineno, col, yytext);) col += yyleng; 
                                   long int value  = strtol(yytext, NULL, 0);
                                   yylval.INTEGER = value;
                                   return yyINTEGER; }
{OCTAL}{SIGNED}/{FOLLOW}         { DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d OCTAL[C|L]         %s\n", lineno, col, yytext);) col += yyleng;
                                   long int value  = strtol(yytext, NULL, 0);
                                   yylval.INTEGER = value;
                                   int token       = yyINTEGER;
                                   if ((yytext[yyleng-1] | 0x20) == 'c') {
                                       token       = yyCHAR;
                                       yylval.CHAR = yylval.INTEGER;
                                   }
                                   return token;
                                 }
{OCTAL}{UNSIGNED}/{FOLLOW}       { DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d OCTAL[UC|UL]       %s\n", lineno, col, yytext);) col += yyleng;
                                   long unsigned int value = strtoul(yytext, NULL, 0);
                                   yylval.UINTEGER  = value;
                                   int token         = yyUINTEGER;
                                   if ((yytext[yyleng-1] | 0x20) == 'c') {
                                       token         = yyUCHAR;
                                       yylval.UCHAR = yylval.UINTEGER;
                                   }
                                   return token;
                                 }
{DECIMAL}/{FOLLOW}               { DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d DECIMAL            %s\n", lineno, col, yytext);) col += yyleng;
                                   yylval.INTEGER = atoi(yytext);
                                   return yyINTEGER; }
{DECIMAL}{SIGNED}/{FOLLOW}       { DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d DECIMAL[C|L]       %s\n", lineno, col, yytext);) col += yyleng;
                                   yylval.INTEGER  = atoi(yytext);
                                   int token        = yyINTEGER;
                                   if ((yytext[yyleng-1] | 0x20) == 'c') {
                                       token        = yyCHAR;
                                       yylval.CHAR = yylval.INTEGER;
                                   }
                                   return token;
                                 }
{DECIMAL}{UNSIGNED}/{FOLLOW}     { DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d DECIMAL[UC|UL]     %s\n", lineno, col, yytext);) col += yyleng;
                                   long unsigned int value = strtoul(yytext, NULL, 0);
                                   yylval.UINTEGER  = value;
                                   int token         = yyUINTEGER;
                                   if ((yytext[yyleng-1] | 0x20) == 'c') {
                                       token         = yyUCHAR;
                                       yylval.UCHAR = yylval.UINTEGER;
                                   }
                                   return token;
                                 }
{HEXADECIMAL}/{FOLLOW}           { DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d HEXADECIMAL        %s\n", lineno, col, yytext);) col += yyleng;
                                   long int value   = strtol(yytext, NULL, 0);
                                   yylval.UINTEGER = value;
                                   return yyINTEGER; }
{HEXADECIMAL}{SIGNED}/{FOLLOW}   { DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d HEXADECIMAL[C|L]   %s\n", lineno, col, yytext);) col += yyleng;
                                   long int value = strtol(yytext, NULL, 0);
                                   yylval.UINTEGER = value;
                                   int token        = yyINTEGER;
                                   if ((yytext[yyleng-1] | 0x20) == 'c') {
                                       token        = yyCHAR;
                                       yylval.CHAR = yylval.INTEGER;
                                   }
                                   return token;
                                 }
{HEXADECIMAL}{UNSIGNED}/{FOLLOW} { DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d HEXADECIMAL[UC|UL] %s\n", lineno, col, yytext);) col += yyleng;
                                   long unsigned int value = strtoul(yytext, NULL, 0);
                                   yylval.UINTEGER  = value;
                                   int token         = yyUINTEGER;
                                   if ((yytext[yyleng-1] | 0x20) == 'c') {
                                       token         = yyUCHAR;
                                       yylval.UCHAR = yylval.UINTEGER;
                                   }
                                   return token;
                                 }
{FLOAT1}/{FOLLOW}                { DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d 1.E1               %s\n", lineno, col, yytext);) col += yyleng;
                                   yylval.FLOAT = atof(yytext);
                                   return yyFLOAT; }
{FLOAT2}/{FOLLOW}                { DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d  .E2               %s\n", lineno, col, yytext);) col += yyleng;
                                   yylval.FLOAT = atof(yytext);
                                   return yyFLOAT; }
{FLOAT3}/{FOLLOW}                { DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d 3E3                %s\n", lineno, col, yytext);) col += yyleng;
                                   yylval.FLOAT = atof(yytext);
                                   return yyFLOAT; }
{HEXFLOAT}/{FOLLOW}              { DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d hexfloat           %s\n", lineno, col, yytext);) col += yyleng;
                                   yylval.FLOAT = ATOF(yytext);
                                   return yyFLOAT; }
{FALSE}/{FOLLOW}                 { DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d FALSE              %s\n", lineno, col, yytext);) col += yyleng;
                                   yylval.BOOL = false;
                                   return yyBOOL; }
{TRUE}/{FOLLOW}                  { DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d TRUE               %s\n", lineno, col, yytext);) col += yyleng;
                                   yylval.BOOL = true;
                                   return yyBOOL; }
{NAME}/{FOLLOW}                  { 
                                   col += yyleng; 
                                   if (!strcmp("include", yytext)) {
                                      DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d include            %s\n", lineno, col, yytext);) 
                                      return yyINCLUDE;
                                   } else if (!strcmp("user", yytext)) {
                                      DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d user               %s\n", lineno, col, yytext);) 
                                      return yyUSER;
                                   } else {
                                      DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d NAME               %s\n", lineno, col, yytext);) 
                                      yylval.STRING = new string(yytext);
                                      return yyNAME;
                                   }
                                 }
.                                { 
                                   /* collect all illegal characters into buffer  */
                                   /* This is the default junk collector. When    */
                                   /* an illegal construct is discovered this     */
                                   /* little jewel puts all illegal characters    */
                                   /* into a buffer, tokenizes it and then ships  */
                                   /* it to the parser.                           */

                                   /**********************************************************************/
                                   /* Character Translation Table: '1' character is not legal.           */
                                   /**********************************************************************/
                                   
                                     static const char xTable[256] =             /* class translation table*/
                                     {
                                   /*
                                   nul  soh  stx  etx  eot  enq  ack  bel  bs   hs   lf   vt   ff   cr   so   si*/
                                     1,   1,   1,   1,   1,   1,   1,   0,   1,   1,  0,   0,   0,   0,   1,   1,
                                   
                                   /*
                                   dle  dc1  dc2  dc3  dc4  nak  syn  etb  can  em   sub  esc  fs   gs   rs   us*/
                                     1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,
                                   
                                   /*
                                   sp    !    "    #    $    %    &    '    (    )    *    +    -    ,    .    /*/
                                     0,   1,   1,   1,   1,   1,   1,   1,   0,   0,   1,   1,   1,   0,   1,   0,
                                   
                                   /*
                                    0    1    2    3    4    5    6    7    8    9    :    ;    <    =    >    ?*/
                                     1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   0,   0,   0,   1,   0,   1,
                                   
                                   /*
                                    @    A    B    C    D    E    F    G    H    I    J    K    L    M    N    O*/
                                     1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,
                                   
                                   /*
                                    P    Q    R    S    T    U    V    W    X    Y    Z    [    \    ]    ^    _*/
                                     1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,
                                   
                                   /*
                                    `    a    b    c    d    e    f    g    h    i    j    k    l    m    n    o*/
                                     1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,
                                   
                                   /*
                                    p    q    r    s    t    u    v    w    x    y    z    {    |    }    ~  del*/
                                     1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   1,   0,   1,   0,   1,   1,
                                   
                                   /* the remaining 128 characters are not part of the legal ASCII character set*/
                                     };                                   
                                   char  buffer[256];
                                   char* ptr = buffer;
                                   *ptr = yytext[0];
                                   do {
                                      if (ptr < &buffer[255])
                                         *(++ptr) = yyinput();
                                      else break;
                                   } while (xTable[(int)*ptr]);
                                   unput(*ptr);
                                   *ptr = 0;
                                   printf("Lex: %s:%d.%d illegal character(s) in file \"%s\"\n"
                                         , getFileName().c_str(), lineno, col, buffer);
                                   col += (ptr - buffer);
                                 }
<INITIAL><<EOF>>                 { return yyEND; }

<COMMENT>{
"*"+                             { DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d COMMENT ***        `%s`\n", lineno, col, yytext);) col += yyleng; }
.*"*/"                           { DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d END COMMENT        `%s`\n", lineno, col, yytext);) col += yyleng; 
                                   BEGIN(INITIAL); }
[^*\n]+                          { DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d COMMENT [^*\\n]    `%s`\n", lineno, col, yytext);) col += yyleng; }
<<EOF>>                          { DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d COMMENT EOL\n", lineno, col);) lineno++; 
                                   BEGIN(INITIAL);
                                   return yyEND; }
\n                               { DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d COMMENT \\n\n", lineno, col);) lineno++; col=1; }
} /* <COMMENT> */


<EOLCOMMENT>{
.**                              { DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d EOL COMMENT        `%s`\n", lineno, col, yytext);) col += yyleng; }
<<EOF>>                          { DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d END COMMENT \\n\n", lineno, col);) lineno++; col=1; 
                                   BEGIN(INITIAL);
                                   return yyEND; }
\n                               { DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d END COMMENT \\n\n", lineno, col);) lineno++; col = 1; 
                                   BEGIN(INITIAL); }
}  /* <EOLCOMMENT> */

<STRING_CHAR>{   
\"/{FOLLOW}                      {  if (stringFlag) {
                                       DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d END STRING         \"%s\"\n", lineno, col, string_buf->c_str());) 
                                       col += yyleng;
                                       BEGIN(INITIAL); 
                                       yylval.STRING = string_buf;
                                       return yySTRING;
                                    } else {
                                       DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d CHARACTER \"        \'%s\'\n", lineno, col, yytext);) 
                                       col += yyleng;
                                       string_buf->append(yytext);
                                    }
                                 }
                                 
\'/{FOLLOW}                      {  if (!stringFlag) {
                                       DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d END CHARACTER      \'%s\'\n", lineno, col, string_buf->c_str());) 
                                       col += yyleng;
                                       BEGIN(INITIAL); 
                                       if (string_buf->size() > 1) {
                                          printf("Lex: %s:%d.%d converting to string \"%s\"\n"
                                                        , getFileName().c_str(), lineno, col, string_buf->c_str());
                                          yylval.STRING = string_buf;
                                          return yySTRING;
                                       } else {
                                          yylval.CHAR = (string_buf->c_str())[0];
                                          delete string_buf;
                                          return yyCHAR;
                                       }
                                    } else {
                                       DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d STRING   \"%s\"\n", lineno, col, yytext);) 
                                       col += yyleng;
                                       string_buf->append(yytext);
                                    }
                                 }
                                 
<<EOF>>                          { DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d END STRING FAIL    \"%s\"\n", lineno, col, string_buf->c_str());) 
                                   col += yyleng;
                                   printf("Lex: %s:%d.%d EOF file before terminal quote \"%s\": ignored\n"
                                                 , getFileName().c_str(), lineno, col, string_buf->c_str());
                                   BEGIN(INITIAL); 
                                   yylval.STRING = string_buf;
                                   return yySTRING;
                                 }
                                 
\n                               { col = 1; lineno++; }

\\0[0-7]{1,3}                    { DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d \\0[0-7]{1,3}       \"%s\"\n", lineno, col, yytext);) 
                                   col += yyleng;
                                   char result[2] = {0, 0};
                                   (void) sscanf( (yytext + 1), "%o", (unsigned int*)&result[0] );
                                   string_buf->append(result);
                                }
\\x[[:xdigit:]]{1,2}            { DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d \\x[[:xdigit:]]{1,2}   \"%s\"\n", lineno, col, yytext);) 
                                   col += yyleng;
                                   char result[2] = {0, 0};
                                   (void) sscanf( (yytext + 2), "%x", (unsigned int*)&result[0] );
                                   string_buf->append(result);
                                 }
     
\\[1-9][[:digit:]]{1,3}          { DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d [1-9][[:digit:]]{1,2}   \"%s\"\n", lineno, col, yytext);) 
                                   col += yyleng;
                                   union {
                                     int32_t result;
                                     char    array[5];
                                   } r;
                                   r.array[4] = 0;
                                   (void) sscanf( (yytext + 1), "%d", &r.result );
                                   if (r.result > 0xFF) {
                                     printf("Lex: %s:%d.%d illegal decimal number in file \"%s\": ignored\n"
                                                   , getFileName().c_str(), lineno, col, yytext);
                                   } else string_buf->append(&r.array[0]);
                                 }
     
\\b                             { DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d STRING \\b            \"%s\"\n", lineno, col, yytext);) col += yyleng;
                                  string_buf->append("\b");
                                }
\\f                             { DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d STRING \\f            \"%s\"\n", lineno, col, yytext);) col += yyleng;
                                  string_buf->append("\f");
                                }
\\n                             { DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d STRING \\n            \"%s\"\n", lineno, col, yytext);) col += yyleng;
                                  string_buf->append("\n");
                                }
\\r                             { DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d STRING \\r            \"%s\"\n", lineno, col, yytext);) col += yyleng;
                                  string_buf->append("\r");
                                }
\\t                             { DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d STRING \\t            \"%s\"\n", lineno, col, yytext);) col += yyleng;
                                  string_buf->append("\t");
                                }
     
\\.                             { DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d STRING \\n         \"%s\"\n", lineno, col, yytext);) col += yyleng;
                                  string_buf->append(yytext);
                                }
     
[^\\\n\"\']+                    { DEBUG(lexerDebugFlag, printf("Lex: %3d:%4d STRING [^\\\\n\"]+    \"%s\"\n", lineno, col, yytext);) col += yyleng;
                                  string_buf->append(yytext);
                                }
} /* <STRING_CHAR> */

%%

# include <string>
using namespace std;

/**
 * @file    Slip.lex
 * @author  A. Schwarz
 * @date    November 16, 2023
 *
 * @brief Lexer regular expressions and integrated classes.
 *
 */

 /**
 * @brief Instantiation of state push/pop functionality
 *
 * <p>The SlipRead class instantiates athe ParseLexState class.
 *    A pointer to the object is contained in state, an external
 *    variable. The intent is to provide a single point for object
 *    creation and manipulation of include file handling. </p>
 */

/**
 * @class ParseLexState
 * @brief Save, manage, and restore lexer state in include file processing.
 *
 * <p>When an include file is seen, the lexer state is saved. When an include
 *    file is exited, the lexer state is restored. The lexer state are those
 *    state variables not managed by the lexer and the input buffer.</p>
 *
 * <p>Sequence of Operations:</p>
 *    1. Until an include file is seen by the parser, the stack is empty.
 *       The Variable object is invalid.
 *    2. When the parser sees an include, a new Variable object is pushed.
 *       onto the stack. The Variable values refer to the state of the lexer
 *       prior to input of the include file contents.
 *    3. When an EOF is seen, the stack is popped, and the popped values
 *       become the current state of the lexer.
 *
 * <p>Indexing starts at zero. The index points to the next entry on the stack.
 *    This implies that an index of zero means the stack is empty.</p>
 */
class ParseLexState {
// #  ifndef YYLTYPE
      // struct YYLTYPE {    \
        // int first_line;   \
        // int first_column; \
        // int last_line;    \
        // int last_column;  \
      // };
// #  endif

  /**
   * @brief Structure maintains the lexer state.
   */
   class Variables {                            //       position 
   public:
      int first_line   = 0;                     //!< lexor line number
      int first_column = 0;                     //!< lexer column number
      int last_line    = 0;                            //!< lexor line number
      int last_column  = 0;                     //!< lexer column number
                                                //       lexer state
      FILE * file      = NULL;                  //!< fopen parameter
      string fileName  = "";                    //!< filename
      
      /*---------------------*
       *    Constructors     *
      *---------------------*/
      /**
       * @brief Constructs a Variables object and opens the data file.
       *
       * <p>Constructing a Variables object is done whenever a new lexer input
       *    buffer is needed. The Variables object contains the locations
       *    required by YYLTYPE, the data structure used by the bison generated
       *    parser, and the file name and descriptor of the current file.</p>
       *
       * @param[in] token    (YYLTYPE) the bison needed roken location data.
       * @param[in] fileName (string)  the name ("path/name") of the file to be opened.
       * @param[in] file     (FILE*)   C file descriptor
       */
      Variables(YYLTYPE& token, string fileName = "", FILE * file = NULL)
               : first_line(token.first_line)
               , first_column(token.first_column)
               , last_line(token.last_column)
               , last_column(token.last_column)
               , fileName(fileName) 
               , file(file) { }
      Variables(string fileName, FILE* file) 
               : first_line(1)
               , first_column(1)
               , last_line(1)
               , last_column(1)
               , fileName(fileName) 
               , file(file) { }
      Variables(void) 
               : first_line(0)
               , first_column(0)
               , last_line(0)
               , last_column(0)
               , fileName("") 
               , file(NULL) { }
      Variables(int first_line, int first_column, int last_line, int last_column, string fileName, FILE* file)
               : first_line(first_line)
               , first_column(first_column)
               , last_line(last_column)
               , last_column(last_column)
               , fileName(fileName) 
               , file(file) { }
      
      /*---------------------*
       *     Functions       *
      *---------------------*/
      
      Variables& operator=(const YYLTYPE& token) {
         first_line   = token.first_line;
         first_column = token.first_column;
         last_line    = token.last_line;
         last_column  = token.last_column;
         return *this;
      }; // Variables& operator=(const YYLTYPE& token) 

      inline string& filename(){ return fileName; }
   }; // class Variables
   
   static const int MAX_ALLOCATION = 64;        //!< Derfault allocation size
   struct Variables * stack;                    //!< Pointer to state stack
   int currentAllocation = MAX_ALLOCATION;      //!< Current size of state stack
   int ndx = 0;                                 //!< Current stack location
   
   string fileName;                             //!< Current file name
   FILE*  file;                                 //!< Current file
   YY_BUFFER_STATE state;                       //!< Current lexer input buffer
   
   public:
      /*---------------------*
       *    Constructors     *
      *---------------------*/
   /**
    * @brief Constructor for state stacking class.
    */
   ParseLexState(string& fileName) 
                    : stack(new Variables[MAX_ALLOCATION]) //!< Allocate initial stack size
   { 
      FILE* file = openfile(fileName);
      stack[ndx]  = Variables(fileName, file);  // stack the first object
      if ( file ) {                             // file successfully opened
         YY_BUFFER_STATE buffer = yy_create_buffer ( file, YY_BUF_SIZE );
         yypush_buffer_state ( buffer );        // stack first lexer input buffer
      }
   }; // ParseLexState(string& fileName)
   
   /*
    * @brief Destructor for state stacking class.
    *
    * <p>The state stack is deleted.</p>
    */
   ~ParseLexState() { deleteState(); }
   
   /*-------------------------*
    *  Begin Stack Operations *
   *--------------------------*/
   /**
    * @brief Return the current state stack depth.
    *
    * <p>The ndx points to the next available slot in the stack array. Indexing
    *    in C/C++ starts at zero, therefore an ndx ==1 => a single entry on the
    *    stack.</p>
    * 
    * @return current depth of stack
    */
   inline int size() { return ndx; }
   
   /**
    * @brief Return the filename of the currently open context.
    *
    * <p>NOTE: No check is made to see if the stack exists of is emply.</p>
    *
    * @return (string) fileName of open context
    */
   inline string& getFileName() { return stack[ndx].filename(); }
      
   /**
    * @brief Pop the last entry in the variable stack and lexer input buffer stack.
    *
    * <p>The bison location data is restored and the lexer input buffer stack is
    *    restored to their values at the time of the push.</p>
    *
    * <p>Push and pop preserves synchronization of the bison and lexer states.</p>
    *
    * @return empty variable object
    *
    */
   inline Variables pop() {
      if (ndx > 0) {
         toString(ndx-1, "pop");
         stack[--ndx];                          // pop stack and update bison location data
         yypop_buffer_state( );                 // pop lexer input buffer
         return stack[ndx];
      }
      return Variables(0, 0, 0, 0, "", NULL);
   };  // void Variables pop()
   
   /**
    * @brief Update the current stack top with bison location data, push a 
    *        new stack entry and create and push a new lexer input buffer.
    *
    * <p>If the current stack is too small to hold another entry then a new 
    *    stack is allocated, all entries are copied and the old stack is deleted.</p>
    *
    * <p>Push and pop preserves synchronization of the bison and lexer states.</p>
    *
    * <p>The sequence of operations:
    *    1. If there is not enough room in the current stack for a new entry:
    *       a. Allocation more memory;
    *       b. Copy the old stack entries to the new stack.
    *       c. Delete the old stack.
    *       d. Substitute the new stack for the old stack.
    *    2. Update the stack top with the latest bison location data.
    *    3. Push a new variable object onto the stack and update the
    *       stack index.
    *    4. Create a new lexer input buffer object.
    *    5. Push the lexer input buffer object onto the lexer stack.
    *    *
    * yylloc              (YYLTYPE&) bison location data
    *
    * @param[in] fileName (string&)  filename of include file
    * @param[in] file     (FILE*)    C descriptor object for file
    * </p>
    */
   void push(string& fileName, FILE* file) {
       if (ndx >= currentAllocation) {
         Variables * localStack = new Variables[currentAllocation + MAX_ALLOCATION];
         int localNdx = 0;
         while(localNdx < currentAllocation) 
            localStack[localNdx] = stack[localNdx++];
         currentAllocation += MAX_ALLOCATION;
         delete stack;
         stack = localStack;
      }
      stack[ndx]         = yylloc;
      stack[++ndx]       = Variables(fileName, file);
      YY_BUFFER_STATE buffer = yy_create_buffer ( file, YY_BUF_SIZE );
      yypush_buffer_state ( buffer );        // stack first lexer input buffer
      toString(ndx, "push");
   };  // void push(YYLTYPE& token, string& fileName, FILE* file)
   
   /*-------------------------*
    *   End Stack Operations  *
   *--------------------------*/
      
      /*---------------------*
       *     Functions       *
      *---------------------*/
      
   /**
    * @brief Open a file.
   */
   inline static
   FILE* openfile(string& filename) {
      FILE* file = fopen(filename.c_str(), "r");
      if (!file) printf("Lexer::saveLexState Open Error on %s:%s\n", filename, strerror(errno));
      return file;
   }; // FILE* openfile(string& filename)
   
   
   inline bool isValid() { return (ndx >= 0)? stack[ndx].file: 0; }
   
   /*-------------------------*
    *  Begin Lexer State Ops  *
   *--------------------------*/
    
   /**
    * @brief Recover the previous lexical state.
    * <p>When an EOF is seen at the end of the lexer input file or
    *    the end of an 'include' file, the previous context is
    *    restored and the currently open file is closed. 
    *    Restoration restores the saved lexer state and
    *    the saved line number and column number.</p>
    *
    * yylloc              (YYLTYPE&) bison location data
    * @return true if restoreState succeeded
    */
   inline
   bool restoreState() {
      Variables var = pop();
      yylloc.first_line   = var.first_line;
      yylloc.first_column = var.first_column;
      yylloc.last_line    = var.last_line;
      yylloc.last_column  = var.last_column;
      return true;
   };  // void restoreState()
       
        
   /**
    * @brief Stack the environment and create a new environment.
    *
    * <p>The input file is opened, the lexer input buffer and
    *    bison token location variables are stacked.</p>
    *
    * yylloc              (YYLTYPE&) bison location data
    *
    * @param[in] fileName (string&)  include file name
    * @return <b>true</b> open of input file successful
    */
   inline bool saveState(string& fileName) {
      FILE* file = openfile(fileName);
      if ( file ) push(fileName, file);
      return (file)? true: false;
   }; // bool saveState(string& fileName)
   
   /**
    * @brief pretty print an entry from the stack.
    *
    * @param[in] ndx (int) stack entry index
    * @param[in] text (string&) title of output
    */
   void toString(const int ndx, const string& text) {
       DEBUG(lexerDebugFlag, printf("Lex: %4s stack[%d], %s:<%3d %3d>\n"
             , text.c_str(), ndx
             , stack[ndx].fileName.c_str()
             , stack[ndx].lineno
             , stack[ndx].col);)
   };  // void toString(const int ndx, const string& text)

/**
 * @brief Cleanup scanner by deleting allocated storage.
 */
   void deleteState() {
      if (stack) delete stack;
      for (; ndx >= 0; ndx--) yypop_buffer_state( );
   };  // void deleteLexState()
}; //  class ParseLexState


   /*-----------------------------*
    *      inteface methods       *
   *------------------------------*/

ParseLexState* state = NULL;                //!< multi-buffer maintenance object

   /**
    * @brief Create the state multi-buffer mainenance object.
    * 
    * The multi-buffer maintenance function creates the object supporting
    * the synchronous pushing/poping of bison and flex data objects when
    * and include file is seen. This simoultaneous operatation is done with
    * direct support of flex v2.6.4 functions and with internal stacks used
    * to push bison objects.
    * 
    * The functionality is not reentrant or recursive. Before each use of
    * input of a new list file, the existing object should be delected 
    * before a new object is created. Not doing this will create havoc, be
    * the genesis of unfavorable opinions by management, and ultimately, 
    * bedlam.
    * 
    * There are several functions performed. (1) a new object is created,  
    * (2) the file containing the list to be read is opened, and (3) the
    * stack is initialized to contain the new bison and flex object.
    * 
    * Some part of the operations are done in the object constructor.
    * 
    * @param[in] filename (string&) 
    * @return true if creation successful
    * 
    */
    bool createState(string& filename) {
       state = new ParseLexState(filename);
       return state->isValid();
    }; // bool createState(string& filename)

/**
 * @brief Cleanup scanner by deleting allocated storage.
 */
void deleteState() {
   if (state) delete state;
};  // void deleteLexState()

/**
 * @brief Get the file name of the current input file.
 * @return (string) file name
 */
string getFileName() {
   return state->getFileName();
};  // string getFileName()

/**
 * @brief Turn debugging on.
*/
void setLexerDebugON() {
   lexerDebugFlag = true;
};  // void setLexerDebugON()

/**
 * @brief Turn debugging off.
*/
void setLexerDebugOFF() {
   lexerDebugFlag = false;
};  // void setLexerDebugOFF()

/**
 * @brief Restore the previous parser/scanner state.
 * <p>Current file replaced by previous file.</p>
 * @note
 * <p>The lineno and col have file scope. They are not part of the 
 *    local scope and their values must be made available to the
 *    lexer. The lexer places the values in file scope and this
 *    little jewel uses it. It is a violation of basic development
 *    principles, but there you have it.</p>
 *
 * @return <b>true</b> restore previous state
 */
bool restoreState() {
   return state->restoreState();
};  // bool restoreState()

/**
 * @brief For each include file, stack the current state.
 *
 * <p>This function accesses methods within the scope of the scanner
 *    file. This access to file scoped variables allows which does
 *    not require any file scoped variables allows usage outside of
 *    the scanner file, most particular in the parser. This function
 *    is the glue which allows the parser to perform lexer-centric
 *    operations.</p>
 *
 * yylloc   (YYLTYPE&) Bison location structure
 *
 * @param[in] filename (string)   include file name
 */
bool saveState(string& fileName) {
   return state->saveState(fileName);
};  // bool saveState(string filename)

   /*-----------------------------*
    *        lexer methods        *
   *------------------------------*/


/**
 * @brief Converts a <i>hexfloat</i> into a double.
 * <p>The conforms to the IEEE 754 layout of double precision 64-bit) floating
 *    point numbers. A hardware or software implementation different from this
 *    standard will cause the algorithm to fail.</p>
 * <p>The algorithm assumes that a 64-bit memory representation is available
 *    in the hardware or through the compiler, and that the 64-bits are
 *    contiguous. Any operation made on the 64-bit memory item assumes that
 *    all 64-bits take part therefore a bit shift will shit within a 64-bit
 *    contiguous memory object.</p>
 * <p>It is assumed tht the input ASCII representation is legal. that is that
 *    it conforms exactly to a valid <i>hexfloat</i> representation. With this
 *    in mind:</p>
 * <ul>
 *    <li>Ignore leading non-graphic ASCII characters (except a <b>NUL</b>.</li>
 *    <li>Check for a signed <i>hexfloat</i>.</li>
 *    <ul>
 *       <li>If the leading character is a '+', ignore and increment over it.</li>
 *       <li>If the leading character is a '-', shift a '<b>1</b>' into the
 *           most signmificat bit of the 64-bit number and increment over the
 *           character.</li>
 *    </ul>
 *    <li>Set the mantissa to zero. For all characters in the mantissa:</li>
 *    <ul>
 *       <li>Convert the character into an integer.</li>
 *       <li>Shift the previous mantissa value by 4-bits.</li>
 *       <li>'or' the current value into the mantissa.</li>
 *       <li>Remember how many bits have been shifted.</li>
 *    </ul>
 *    <li>If the number of bits shifted is less than the allowable maximum,
 *        then shift the mantissa. If the number exceed the maximum, ignore,
 *        the returned value will not be legal.</li>
 *    <li>Process the exponent:</li>
 *    <ul>
 *       <li>Skip of the exponent indicator, '<b>p</b>'.</li>
 *       <li>Remember the exponsent sign (+, -).</li>
 *       <li>Construct the exponen valuet. The exponent is a decimal number,
 *           -1023 &lt; x &lt; 1023. </li>
 *       <li>Multiply by the sign, add the bias, and shift.</li>
 *       <li>'or' the exponent in to the double precision integer cognate.</li>
 *    </ul>
 * </ul>
 * @param str (char *) <i>hexfloat</i> string "0x1.#p#"
 * @return floating point value
 */
double myatof(char * str) {

      /**********************************************************************/
      /* Character Translation Table: '255' character is not legal.         */
      /**********************************************************************/

        static const uint8_t xTable[256] =             /* class translation table*/
        {
      /*
      nul  soh  stx  etx  eot  enq  ack  bel  bs   hs   lf   vt   ff   cr   so   si*/
      255, 255, 255, 255, 255, 255, 255, 255, 255, 255,255, 255, 255, 255, 255, 255,

      /*
      dle  dc1  dc2  dc3  dc4  nak  syn  etb  can  em   sub  esc  fs   gs   rs   us*/
      255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255,

      /*
      sp    !    "    #    $    %    &    '    (    )    *    +    -    ,    .    /*/
      255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255,

      /*
       0    1    2    3    4    5    6    7    8    9    :    ;    <    =    >    ?*/
        0,   1,   2,   3,   4,   5,   6,   7,   8,   9, 255, 255, 255, 255, 255, 255,

      /*
       @    A    B    C    D    E    F    G    H    I    J    K    L    M    N    O*/
      255,  10,  11,  12,  13,  14,  15, 255, 255, 255, 255, 255, 255, 255, 255, 255,

      /*
       P    Q    R    S    T    U    V    W    X    Y    Z    [    \    ]    ^    _*/
      255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255,

      /*
       `    a    b    c    d    e    f    g    h    i    j    k    l    m    n    o*/
      255,  10,  11,  12,  13,  14,  15, 255, 255, 255, 255, 255, 255, 255, 255, 255,

      /*
       p    q    r    s    t    u    v    w    x    y    z    {    |    }    ~  del*/
      255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255,

      /* the remaining 128 characters are not part of the legal ASCII character set*/
      };
   static const int MANTISSA_SHIFT =    52;
   static const int EXP_MASK       = 0x7ff;
   static const int EXP_BIAS       = 0x3ff;
   union {
      double     dbl;
      uint64_t dblDef;
   } result;
   uint64_t mantissa = 0;
   int      shift    = 0;
   int      digit    = 0;
   char*    p        = str;
   result.dbl = 0.0;
   while((*p > 0) && (*p <= (uint8_t)' ')) p++;
   if (*p != '0') {
      if (*p == '-')
         result.dblDef = (uint64_t)1 << 63;
      p++;
   }
   p += 4;

   while((digit = xTable[(uint8_t)*p]) != 255) {
      mantissa = (mantissa << 4) | digit;
      shift += 4;
      p++;
   }
   if (shift < MANTISSA_SHIFT) mantissa = mantissa << (MANTISSA_SHIFT - shift);
   result.dblDef |= mantissa;
   if (*p == 'p' || *p == 'P') {
      uint64_t exp = 0;
      p++;
      int sign = 1;
      if (*p == '+') p++;
      else if (*p == '-') {
         sign = -1;
         p++;
      }
      while((digit = xTable[(uint8_t)*p]) != 255) {
         exp = exp*10 + digit;
         p++;
      }
      exp   = sign*exp + EXP_BIAS;
      exp  &= EXP_MASK;
      exp <<= MANTISSA_SHIFT;
      result.dblDef |= exp;
   }
   return result.dbl;
};  // double myatof(char * str)
