// Copyright (C) 2014 Arthur I. Schwarz
//
// This file is part of the C++ SLIP library.  This library is free
// software; you can redistribute it and/or modify it under the
// terms of the GNU General Public License as published by the
// Free Software Foundation; either version 3, or (at your option)
// any later version.

// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// Under Section 7 of GPL version 3, you are granted additional
// permissions described in the GCC Runtime Library Exception, version
// 3.1, as published by the Free Software Foundation.

// You should have received a copy of the GNU General Public License and
// a copy of the GCC Runtime Library Exception along with this program;
// see the files COPYING3 and COPYING.RUNTIME respectively.  If not, see
// <http://www.gnu.org/licenses/>.

/**
 * @brief Description of the SLIP grammar.
 * <pre><tt>
 * Syntax and tokens for SlipReadList.lex.cpp
 * Sublist Declaration
 *                                                 // list Definitions 
 * NAME( {mark} <Descriptor list> ( ) );           // general list definition, insert mark and Descriptor list
 * NAME( <Descriptor list> ( ) );                  // list Definition insert Descriptor list
 * NAME( {mark} ( ) );                             // list Definition insert user mark
 * NAME( );                                        // list Definition of an empty list
 * NAME( {mark} <Descriptor list> {NAME};          // expand NAME and insert mark and Descriptor list
 * NAME( {mark} {listNAME} );                      // expand listNAME and insert mark
 * NAME( <Descriptor list> {listNAME} );           // expand listNAME and insert Descriptor list
 *
 * User Data Declaration
 * user userName;                                  // Single user data name
 * user userName, ...;                             // sublist of user data names
 * userName{ (list) }                              // User Data converted to SlipDatum(PTR);
 *
 * #include "path/name"                            // Include file
 *
 *  Syntax                                         // list Stuff
 *
 * 'c'                                             // char           SlipDatum((char)'c')
 * c                                               // char           SlipDatum((char)'c')
 * "string"                                        // string         SlipDatum((string)string, false)
 * true                                            // bool           SlipDatum((bool)true)
 * false                                           // bool           SlipDatum((bool)true)
 * numberUC                                        // unsigned char  SlipDatum((UCHAR)number)
 * numberC                                         // signed char    SlipDatum((CHAR)number)
 * numberUL                                        // unsigned long  SlipDatum((ULONG)number)
 * numberL                                         // signed long    SlipDatum((LONG)number)
 * number                                          // signed long    SlipDatum((LONG)number)
 * double        1.0E+0                            // double         SlipDatum(DOUBLE)number)
 * double        1.E+0                             // double         SlipDatum(DOUBLE)number)
 * double        1.                                // double         SlipDatum(DOUBLE)number)
 * double         .1E+0                            // double         SlipDatum(DOUBLE)number)
 * double         .1                               // double         SlipDatum(DOUBLE)number)
 * double        1E+0                              // double         SlipDatum(DOUBLE)number)
 * ( {mark} <Descriptor list> ( ) )                // Anonymous list SlipHeader()
 * ( <Descriptor list>( ) )                        // Anonymous list SlipHeader()
 * ( {mark} ( ) )                                  // Anonymous list SlipHeader()
 *
 * {NAME}                                          // list expansion SlipSublist(listName)
 *
 * Comments
 * / *   * /                                      // 'C' comments
 * //                                             // 'C++' comments
 *
 * Special Symbols  '(' ')' '{' '}' ';' '/' '<' '>' '\"'
 * Keywords         include user
 *
 * EXAMPLE
 *
 * list mylist;                         // forward declaration of mylist
 *
 * #include "myFileName"                // include the file "myFileName" 
 *
 * list list1, list2;                   // list of forward declarations
 *
 * user myData;                         // name of a User Data class
 *
 * list3 ( {0x15} < 0 (1 2 {list1}) '3' 4 > () );
 *         |      | | |             |   |    |
 *         |      | | |             |   |    . anonymous sublist part of list3
 *         |      | | |             |   |
 *         |      | | |             |   . value
 *         |      | | |             |
 *         |      | | |             . key
 *         |      | | |
 *         |      | | . anonymous list value in <key v>
 *         |      | |
 *         |      | . key in <k v>
 *         |      |
 *         |      . Descriptor list
 *         |
 *         . user mark for list3
 *
 * list4 ( < "string" { 0x22 (5)} > 9 10 11)
 *         | |        | |    |       |
 *         | |        | |    |       . list items in list4
 *         | |        | |    |
 *         | |        | |    . anonymous list
 *         | |        | |
 *         | |        | . user mark in anonymous list
 *         | |        |
 *         | |        . anonymous list value in <key v>
 *         | |
 *         | . key
 *         |
 *         . Descriptor list <key v> pairs
 *
 * list5 ( key1 value1 key2 value2 )
 *
 * list6 ( <{list5}> "one" 1.3 5UC 6L)    
 *         | |         |
 *         | |         . start of list items
 *         | |
 *         | . Descriptor defined by a list of <key, value> pairs
 *         |
 *         . list
 * 
 * list7 { myData( 3 4 ) };
 *         |     |
 *         |     . list containing user data
 *         |
 *         . user data name
 *
 * list8 { 230 ( able ( baker charlie (foxtrot) one {list1} ) ) };
 *
 * ( this "is" a "list" {list1} (1 2 3) {list5} {10 <1 2> (able)} ) // returned list
 *
 * </tt></pre>
 */
 
 /*
  * Maintenance:
  * Compile: bison -d -t --no-lines --warnings=all --language=c++ --report=all slip.y
  *          cp *.hh ../slip/; cp Slip.Bison.cpp ../slip/; cp Slip.Bison.hpp ../slip/;
 */
 
 /*************************************
  *    Declarations and Options
  *************************************/
 
%code requires {
   # include <stdint.h>
   # include <string>
   # include "SlipCell.h"
   # include "SlipDatum.h"
   # include "SlipDescription.h"
   # include "SlipHeader.h"
   # include "SlipInputMethods.h"
   # include "SlipSublist.h"
 
   # undef BOOL
   # undef CHAR
   # undef FLOAT
   # undef STRING
   # undef UCHAR
   # undef UINTEGER
   # undef INTEGER
   # undef SLIPCELL
   # undef SLIPDATUM
   # undef SLIPHEADER
   # undef SLIPSUBLIST
 
   using namespace std;
}

/*************************************************/
/*                Parser Options                 */
/*************************************************/
 %require "3.8.2"                              // minimum Bison version which supports options
 
 %defines                                      // Generate I/F files (bison -d)
 %no-lines                                     // Suppress #line in generated files
 %header "SlipParser.h"                        // Output header file named "SlipParser.h"////////////////
 %output "SlipParser.cpp"                      // Output parser file named "SlipParser.cpp"
 
 %locations                                    // generate location information
 %defines                                      // Generate I/F files (bison -d)
 %define parse.error detailed                  // Detailed error message passed to yyerror
 %define parse.lac full                        // Lookahead Correction (LAC) - better errors
 %no-lines                                     // Suppress #line in generated files

 %parse-param {SlipHeader** head}              // Parser return value (pointer to a list)
 %parse-param {SlipInputMethods* methods}      // Methods used by parser (from Slip)
 
 %output "SlipParser.cpp"                      // Output Slip.Bison files
 
/*************************************************/
/*              Discarded Options                */
/*************************************************/

/*%language c++                                 // Generate C++ parser*/
/*%name-prefix "Slip"                           // Change 'yy' prefix to Slip*/
/*%define api.namespace { slip }                // Namespace for parser*/
/*%define api.value.type { token_type }         // define YYSTYPE*/
/*%debug                                        // Enable runtime debug information*/

 %{                                             // Insert into .cpp file
   # include "SlipFlex.h"                       // Scanner header file
   # include "SlipParser.h"
   # include "SlipScanner.h"
    
   void yyerror (const char *s);
   void yyerror (SlipHeader** head, SlipInputMethods* methods, const char * msg);
   int yylex();
   
   using namespace std;
 %}
 
%code requires {

#ifdef DEBUG_IO
   #undef DEBUG
   #define DEBUG(flag, stmt) if (flag) { stmt }
#else
   #define DEBUG(flag, stmt)
#endif
}

# ifdef __cplusplus
#   define isatty(arg1) false
# endif


%union YYSTYPE {                               //!< Token values
   bool              BOOL;                     //!< token boolean true / false
   int8_t            CHAR;                     //!< 8-bit  token value
   double            FLOAT;                    //!< 64-bit token value
   int32_t           INTEGER;                  //!< 32-bit token value
   string*           STRING;                   //!< pointer to new string()
   uint8_t           UCHAR;                    //!< 8-bit unsigned value
   uint32_t          UINTEGER;                 //!< 32-bit unsigned value
   SlipCell*         SLIPCELL;                 //!< general slip cell
   SlipDatum*        SLIPDATUM;                //!< data carrying SLIP cell
   SlipHeader*       SLIPHEADER;               //!< list header
   SlipSublist*      SLIPSUBLIST;              //!< list sublist
   SlipDescription*  DESCRIPTION;              //!< descriptionList data passage
}

 %start sublistDeclarations_Definition
 
 // Types associated with terminals
 %token <BOOL>          yyBOOL                 //!< boolean value TRUE or FALSE
 %token <CHAR>          yyCHAR                 //!< a character 'c'
 %token <FLOAT>         yyFLOAT                //!< double value
 %token <INTEGER>       yyINTEGER              //!< a signed long integer
 %token <STRING>        yySTRING               //!< a string "string"
 %token <UCHAR>         yyUCHAR                //!< an unsigned character 'c'
 %token <UINTEGER>      yyUINTEGER             //!< an unsigned long integer
 
 // Constructed terminals
 %token <CHAR>          yyCHARS                //!< a signed 8-bit number
 %token <UCHAR>         yyCHARU                //!< an unsigned 8-bit number
 %token <INTEGER>       yyEND                  //!< End-Of-File indication
 %token <STRING>        yyNAME                 //!< a string [a-zA-Z_$][a-zA-Z0-9_$]*
 
 // Literals: delimiters
 %token <CHAR>          yyCOLON                //!< the literal ':'
 %token <CHAR>          yyCOMMA                //!< the literal ','
 %token <CHAR>          yyHASH                 //!< the literal '#'
 %token <CHAR>          yyPERIOD               //!< the literal '.'
 %token <CHAR>          yySEMI                 //!< the literal ';'
 
 // Literals: multi-character
 %token <INTEGER>       yyINCLUDE              //!< the literal 'include'
 %token <INTEGER>       yyLIST                 //!< the literal 'list'
 %token <INTEGER>       yyUSER                 //!< the literal 'user'
 
 // Literals: balanced delimiters
 %token <CHAR>          yyLPARN                //!< the literal '('
 %token <CHAR>          yyRPARN                //!< the literal ')'
 %token <CHAR>          yyLBRACE               //!< the literal '{'
 %token <CHAR>          yyRBRACE               //!< the literal '}'
 %token <CHAR>          yyLANGLE               //!< the literal '<'
 %token <CHAR>          yyRANGLE               //!< the literal '>'
 
 // Literals: special characters
 %token <CHAR>          yyBELL                 //!< audible bell
 %token <CHAR>          yyBACK                 //!< backspace
 %token <CHAR>          yyFF                   //!< form feed
 %token <CHAR>          yyLF                   //!< line feed
 %token <CHAR>          yyCR                   //!< carriage return
 %token <CHAR>          yyHT                   //!< horizontal tabs


// Types associated with non-terminals
 %type <DESCRIPTION>    datum
 %type <DESCRIPTION>    description
 %type <INTEGER>        declarations
 %type <DESCRIPTION>    descriptionElement
 %type <INTEGER>        declarationItem
 %type <DESCRIPTION>    descriptionItemlist
 %type <DESCRIPTION>    descriptionList
 %type <DESCRIPTION>    namedList
 %type <DESCRIPTION>    listItem
 %type <DESCRIPTION>    number
 %type <DESCRIPTION>    list
 %type <DESCRIPTION>    listDefinition
 %type <DESCRIPTION>    listItemList
 %type <SLIPHEADER>     sublistDeclarations_Definition
 %type <DESCRIPTION>    userData
 %type <INTEGER>        include
 %type <UCHAR>          mark
 %type <STRING>         name
 %type <INTEGER>        userDataList
 
// destruct on errors
%destructor {  methods->incErrorCount();
               $$->deleteData(); 
               $$ = NULL;
            }  datum description descriptionElement 
               descriptionItemlist descriptionList
               namedList listItem number list listDefinition
               listItemList userData

 %%
 /*************************************
  *              BNF
  *************************************/
sublistDeclarations_Definition:   
            declarations listDefinition yyEND    /* list defined, we are done         */
                     {  *head = $$ = (SlipHeader*)($2->getPtr()); 
                        DEBUG(methods->getInputFlag(), methods->debugOut("declarations listDefinition yyEND", $2);)
                        delete $2;
                        YYACCEPT; 
                     }
          | declarations namedList yyEND         /* list defined, we are done         */
                     {  *head = $$ = (SlipHeader*)($2->getPtr()); 
                        DEBUG(methods->getInputFlag(), methods->debugOut("declarations namedList yyEND", $2);)
                        delete $2;
                        YYACCEPT; 
                     }
          | declarations                       /* declaration section               */
                     {  $$ = NULL; 
                        DEBUG(methods->getInputFlag(), methods->debugOut("declarations yyEND", NULL);)
                     }
          | listDefinition yyEND                 /* list defined, we are done         */
                     {  *head = $$ = (SlipHeader*)($1->getPtr()); 
                        DEBUG(methods->getInputFlag(), methods->debugOut("listDefinition yyEND", $1);)
                        delete $1;
                        YYACCEPT; 
                     }
          | namedList yyEND                      /* list defined, we are done         */
                     {  *head = $$ = (SlipHeader*)($1->getPtr());
                        DEBUG(methods->getInputFlag(), methods->debugOut("namedList yyEND", $1);)
                        delete $1;
                        YYACCEPT; 
                     }
          ;

declarations:       
            declarations declarationItem       /* list of declarations              */
                     { DEBUG(methods->getInputFlag(), methods->debugOut("declarations declarationItem", NULL);)
                     }
          | declarationItem                    /* single declaration                */
                     { DEBUG(methods->getInputFlag(), methods->debugOut("declarationItem", NULL);)
                     }
          ;

declarationItem:
            name listDefinition yySEMI         /* list definition                   */
                     {  $$ = 0;
                        methods->declarationItem("name listDefinition", $1, $2);
                     }
          | yyUSER userDataList yySEMI         /* user data names                   */
          | include                            /* include file def'n                */
          | yyEND                              /* EOF check if in an include file   */
                     {  restoreState(); 
                        DEBUG(methods->getInputFlag(), methods->debugOut("yyEND", NULL);)
                     }
          ;

listDefinition:   
            yyLPARN mark description list yyRPARN  /* full list declaration              */
                     {  $$ = methods->markDescriptionList("(mark description list)", $2, $3, $4);
                     }
          | yyLPARN mark description yyRPARN       /* mark and description w/o list      */
                     {  $$ = methods->markDescription("(mark description)", $2, $3);
                     }
          | yyLPARN description list yyRPARN       /* no mark in list                    */
                     {  $$ = methods->descriptionList("(description list)", $2, $3);
                     }
          | yyLPARN description yyRPARN            /* no list just a description         */
                     {  $$ = methods->descriptionNULLList("(description)", $2);
                     }
          | yyLPARN mark list yyRPARN           /* no descriptor in list              */
                     {  $$ = methods->markList("(mark list)", $2, $3);
                     }
          | yyLPARN mark yyRPARN                /* list contains only a mark          */
                     {  $$ = methods->mark("(mark)", $2);
                     }
                                                /* list declaration                  */
          | yyLPARN list yyRPARN                /* list declaration w/o description list */
                     {  $$ = methods->listDefinition("(mark list)", $2);
                     }
          | yyLPARN yyRPARN                     /* empty list                        */
                     {  $$ = methods->nullList("(list)");
                     }
          ;

description:   
           yyLANGLE description descriptionList yyRANGLE /* < < > 1 2>                 */
                     {  $$ = methods->descriptionDescriptionList("<description descriptionList>", $2, $3);
                     }
         | yyLANGLE descriptionList yyRANGLE    /* Association list               */
                     {  $$ = methods->simpleReturn("< descriptionList >", $2);
                     }
          | yyLANGLE description yyRANGLE
                     {  $$ = methods->description("< description >", $2);
                     }
          | yyLANGLE mark yyRANGLE              /* almost empty association list  */
                     {  $$ = methods->markDescription("< mark > ", $2);
                     }
          | yyLANGLE yyRANGLE                   /* empty association list         */
                     {  $$ = methods->nullDescription("< mark > ");
                     }
          ;
          
descriptionList: 
             descriptionItemlist
                     {  $$ = methods->simpleReturn("descriptionItemlist", $1);
                     }
          |  mark descriptionItemlist
                     {  $$ = methods->markDescriptionItemList("mark descriptionItemlist", $1, $2);
                     }
          | namedList
                     {  $$ = methods->simpleReturn("namedList", $1);
                     }
          ;
descriptionItemlist:
                                               /* descriptionItemlist          */
            descriptionItemlist descriptionElement descriptionElement
                     {  $$ = methods->descriptionItemlist("descriptionItemlist", $1, $2, $3);
                     }
                                               /* description item               */
          | descriptionElement descriptionElement
                     {  $$ = methods->descriptionItem("descriptionElement", $1, $2);
                     }
          ;  

descriptionElement:
            namedList                          /* {NAME} list expansion          */
                     {  $$ = $1; }
          | datum                              /* SlipDatum()                    */
                     {  $$ = $1; }
          | yyLPARN listItemList yyRPARN       /* SlipDatum() or SlipHeader      */
                     {  $$ = $2; }
          ;

include:    yyHASH yyINCLUDE datum             /* include file path and name     */
                     {  $$ = 0;
                        methods->includeFile("INCLUDE datum", $3);
                     }
          ;
          
userDataList:                                  /* list of User Data names        */
            userDataList yyCOMMA name
                     {  methods->userDataList("", $3);
                     }
          | name
                     {  methods->userDataList("", $1);
                     }
          ;

list:
            listItemList
                     {  $$ = methods->simpleReturn("list", $1);
                     }
            ;
          
listItemList:
            listItemList listItem              /* listHeader.enqueue(listItem)   */
                     {  $$ = methods->listItemList("listItemList listItem", $1, $2);
                     }
          | listItem                           /* listHeader.enqueue(listItem)   */
                     {  $$ = methods->listItem("listItem", $1);
                     }
          ;

listItem:   
            datum                              /* SlipDatum(datum)               */
                     {  $$ = $1; }
          | namedList                          /* SlipSublist(namedList)         */
                     {  $$ = $1; }
          | userData                           /* SlipDatum(PTR)                 */
                     {  $$ = $1; }
          | listDefinition                     /* SlipSubsublist(list)           */
                     {  $$ = methods->listItemListDefinition("listDefinition", $1);
                     }
          ;

mark:                                          /* list user mark                 */
          yyLBRACE number yyRBRACE
                     {  $$ = (CHAR)(*(SlipDatum*)($2->getPtr()));
                        delete $2;
                     }
          ;
          
namedList: yyLBRACE name yyRBRACE              /* SlipSubsublist(name)           */
                     {  $$ = methods->namedList("{ name }", $2);
                     }
          ;

userData:   name  listDefinition               /* SlipDatum(PTR)                 */
                       { $$ = methods->userData("name listDefinition", $1, $2); }
          | name  namedList
                       { $$ = methods->userData("name {namedList}", $1, $2); }
          ;

datum:      yyBOOL                               /* SlipDatum((bool)boolValue)     */
                     { $$ = new SlipDescription(new SlipDatum($1), SlipDescription::DATA, methods->getDebugFlag()); }
          | yyCHAR                               /* SlipDatum((char)chr)           */
                     { $$ = new SlipDescription(new SlipDatum($1), SlipDescription::DATA, methods->getDebugFlag()); }
          | yyUCHAR                              /* SlipDatum((unsigned char)unsignedChr)*/
                     { $$ = new SlipDescription(new SlipDatum($1), SlipDescription::DATA, methods->getDebugFlag()); }
          | number                             /* discrete number                */
                     { $$ = $1; }
          | yyFLOAT                            /* SlipDatum((double)floatNum)    */
                     { $$ = new SlipDescription(new SlipDatum($1), SlipDescription::DATA, methods->getDebugFlag()); }
          | yySTRING                           /* SlipDatum((string)string)      */
                     { $$ = new SlipDescription(new SlipDatum($1), SlipDescription::DATA, methods->getDebugFlag()); }
          ;
          
number:     yyINTEGER                         /* SlipDatum((long)integer)          */
                     { $$ = new SlipDescription(new SlipDatum($1), SlipDescription::DATA, methods->getDebugFlag()); }
          | yyUINTEGER                        /* SlipDatum((unsigned long)unsignedInteger)*/
                     { $$ = new SlipDescription(new SlipDatum($1), SlipDescription::DATA, methods->getDebugFlag()); }
          | yyCHARS                           /* SlipDatum((char)chr)              */
                     { $$ = new SlipDescription(new SlipDatum($1), SlipDescription::DATA, methods->getDebugFlag()); }
          | yyCHARU                           /* SlipDatum((unsigned char)unsignedChr)*/
                     { $$ = new SlipDescription(new SlipDatum($1), SlipDescription::DATA, methods->getDebugFlag()); }
          ;
 name     : yyNAME
                     { $$ = $1; }
          ;
%%


void error(const YYLTYPE& yylloc, const std::string& msg) {
   fprintf(stderr, "<%4d:%4d> %s\n"
                 , yylloc.first_line
                 , yylloc.first_column
                 , msg.c_str());
}; // error(const location_type& loc, const std::string& msg)

/**
   @brief Standard error reporting for Bison.
*/
void yyerror (const char *s) {
   fprintf (stderr, "%s\n", s);
}; // void yyerror (const char *s)

/**
   @brief Standard error reporting for Bison.
   
   NOTE: This function is not listed in Bison 3.8.1 documentation.
*/
void yyerror (SlipHeader** head, SlipInputMethods* methods, const char * msg) {
   yyerror(msg);
}; // void yyerror (SlipHeader** head, SlipInputMethods* methods, const char * msg)

/**
   @brief Interface function to SlipFlex scanner.
   
   Keeping yylex() defined within the parser allows access to all the
   parser objects. This makes the interface easier in that whatever
   Bison does to allow passage of parser owned objects is bypassed
   in favor of a customized interface. 
   
   The manuals for both bison and Flex are tortuous to read and
   understand. Lord knows why they were written. The Flex 2.6.4 manual
   does not track the Flex 2.6.4 implementation. I have no idea if
   this negligence is recaspitulated in the Bison manual, and don't 
   care. This is the workaround. I can't spend the rest of my life
   fixing interfaces and etc. that are not documented. 
   
   My path forward is to change this poor excuse for a code generator
   to another which is better documented and easier to use. 
   
   Bison Variables Used:
   1. yylval: value of the returned token. yylval is generated by
      %union, therefore, use yylval.<name>.
   2. yylloc: token location given as<br>
      &nbsp;&nbsp;&nbsp;token begin line number
      &nbsp;&nbsp;&nbsp;token begin column number
      &nbsp;&nbsp;&nbsp;token end line number
      &nbsp;&nbsp;&nbsp;token end column number
      
*/
int yylex() {
   return Sliplex();
}; // 
